import tensorflow as tf
from baconian.core.core import Basic, EnvSpec
import numpy as np
import abc
from baconian.core.parameters import Parameters
from typeguard import typechecked
from tensorflow.python.ops.parallel_for.gradients import batch_jacobian as tf_batch_jacobian
from baconian.common.logging import Recorder
from baconian.core.status import register_counter_info_to_status_decorator, StatusWithSingleInfo
from baconian.common.logging import ConsoleLogger
from baconian.common.error import *
from baconian.core.core import EnvSpec, Env
from baconian.algo.dynamics.reward_func.reward_func import RewardFunc
from baconian.algo.dynamics.terminal_func.terminal_func import TerminalFunc
from baconian.common.data_pre_processing import DataScaler, IdenticalDataScaler


class DynamicsModel(Basic):
    STATUS_LIST = ('CREATED', 'INITED')
    INIT_STATUS = 'CREATED'

    def __init__(self, env_spec: EnvSpec, parameters: Parameters = None, init_state=None, name='dynamics_model',
                 state_input_scaler: DataScaler = None,
                 action_input_scaler: DataScaler = None,
                 state_output_scaler: DataScaler = None):

        """

        :param env_spec: environment specifications, such as observation space and action space
        :type env_spec: EnvSpec
        :param parameters: parameters
        :type parameters: Parameters
        :param init_state: initial state of dymamics model
        :type init_state: str
        :param name: name of instance, 'dynamics_model' by default
        :type name: str
        :param state_input_scaler: data preprocessing scaler of state input
        :type state_input_scaler: DataScaler
        :param action_input_scaler: data preprocessing scaler of action input
        :type action_input_scaler: DataScaler
        :param state_output_scaler: data preprocessing scaler of state output
        :type state_output_scaler: DataScaler
        """
        super().__init__(name=name)
        self.env_spec = env_spec
        self.state = init_state
        self.parameters = parameters
        self.state_input = None
        self.action_input = None
        self.new_state_output = None
        self.recorder = Recorder(flush_by_split_status=False, default_obj=self)
        self._status = StatusWithSingleInfo(obj=self)
        self.state_input_scaler = state_input_scaler if state_input_scaler else IdenticalDataScaler(
            dims=env_spec.flat_obs_dim)
        self.action_input_scaler = action_input_scaler if action_input_scaler else IdenticalDataScaler(
            dims=env_spec.flat_action_dim)
        self.state_output_scaler = state_output_scaler if state_output_scaler else IdenticalDataScaler(
            dims=env_spec.flat_obs_dim)

    def init(self, *args, **kwargs):
        self.set_status('INITED')
        self.state = self.env_spec.obs_space.sample()

    @register_counter_info_to_status_decorator(increment=1, info_key='step_counter')
    def step(self, action: np.ndarray, state=None, allow_clip=False, **kwargs_for_transit):

        """
        State transition function (only support one sample transition instead of batch data)

        :param action: action to be taken
        :type action: np.ndarray
        :param state: current state, if None, will use stored state (saved from last transition)
        :type state: np.ndarray
        :param allow_clip: allow clip of observation space, default False
        :type allow_clip: bool
        :param kwargs_for_transit: extra kwargs for calling the _state_transit, this is typically related to the
                                    specific mode you used
        :type kwargs_for_transit:
        :return: new state after step
        :rtype: np.ndarray
        """
        state = np.array(state).reshape(self.env_spec.obs_shape) if state is not None else self.state
        action = action.reshape(self.env_spec.action_shape)
        if allow_clip is True:
            if state is not None:
                    state = self.env_spec.obs_space.clip(state)
            action = self.env_spec.action_space.clip(action)
        if self.env_spec.action_space.contains(action) is False:
            raise StateOrActionOutOfBoundError(
                'action {} out of bound of {}'.format(action, self.env_spec.action_space.bound()))
        if self.env_spec.obs_space.contains(state) is False:
            raise StateOrActionOutOfBoundError(
                'state {} out of bound of {}'.format(state, self.env_spec.obs_space.bound()))
        new_state = self._state_transit(state=state, action=self.env_spec.flat_action(action),
                                        **kwargs_for_transit)
        if allow_clip is True:
            new_state = self.env_spec.obs_space.clip(new_state)
        if self.env_spec.obs_space.contains(new_state) is False:
            raise StateOrActionOutOfBoundError(
                'new state {} out of bound of {}'.format(new_state, self.env_spec.obs_space.bound()))
        self.state = new_state
        return new_state

    @abc.abstractmethod
    def _state_transit(self, state, action, **kwargs) -> np.ndarray:
        """

        :param state: original state
        :type state: np.ndarray
        :param action:  action taken by agent
        :type action: np.ndarray
        :param kwargs:
        :type kwargs:
        :return: new state after transition
        :rtype: np.ndarray
        """
        raise NotImplementedError

    def copy_from(self, obj) -> bool:
        """

        :param obj: object to copy from
        :type obj:
        :return: True if successful else raise an error
        :rtype: bool
        """
        if not isinstance(obj, type(self)):
            raise TypeError('Wrong type of obj %s to be copied, which should be %s' % (type(obj), type(self)))
        return True

    def make_copy(self):
        """ Make a copy of parameters and environment specifications."""
        raise NotImplementedError

    def reset_state(self, state=None):
        """

        :param state: original state
        :type state: np.ndarray
        :return: a random sample space in observation space
        :rtype: np.ndarray
        """
        if state is not None:
            assert self.env_spec.obs_space.contains(state)
            self.state = state
        else:
            self.state = self.env_spec.obs_space.sample()

    def return_as_env(self) -> Env:
        """

        :return: an environment with this dynamics model
        :rtype: DynamicsEnvWrapper
        """
        return DynamicsEnvWrapper(dynamics=self,
                                  name=self._name + '_env')


class LocalDyanmicsModel(DynamicsModel):
    pass


class GlobalDynamicsModel(DynamicsModel):
    pass


class TrainableDyanmicsModel(object):
    def train(self, *args, **kwargs):
        raise NotImplementedError


class DifferentiableDynamics(object):
    @typechecked
    def __init__(self, input_node_dict: dict, output_node_dict: dict):
        for node in input_node_dict.values():
            if not isinstance(node, tf.Tensor):
                raise TypeError('Derivable only support tf.Tensor as node')
        for node in output_node_dict.values():
            if not isinstance(node, tf.Tensor):
                raise TypeError('Derivable only support tf.Tensor as node')
        self.input_node_dict = input_node_dict
        self.output_node_dict = output_node_dict
        self.output_node_list = []
        for key in output_node_dict.keys():
            self.output_node_list.append(output_node_dict[key])
        self._grad_dict = [{}, {}, {}]
        for val in input_node_dict:
            self._grad_dict[0][val] = self.output_node_list

    def grad_on_input(self, key_or_node: (str, tf.Tensor), order=1, batch_flag=False):
        if batch_flag:
            raise NotImplementedError
        node = key_or_node if isinstance(key_or_node, tf.Tensor) else self.input_node_dict[key_or_node]
        if node not in self._grad_dict:
            if order == 1:
                grad_op = [tf_batch_jacobian(output=o_node, inp=node) for o_node in self.output_node_list]
            else:
                grad_op = [self.split_and_hessian(out_node=o_node, innode=node) for o_node in self.output_node_list]
            self._grad_dict[order][node] = grad_op
            return grad_op
        else:
            return self._grad_dict[order][node]

    def split_and_hessian(self, out_node, innode):
        out_nodes = tf.split(out_node, 1, axis=1)
        hessian_node = []
        for o_node in out_nodes:
            hessian_node.append(tf.stack(tf.hessians(o_node, innode)))
        new_dim = len(hessian_node[0].shape.as_list()) + 1
        new_dim = list(range(new_dim))
        new_dim[0] = 1
        new_dim[1] = 0
        return tf.transpose(tf.stack(hessian_node), perm=new_dim)


class DynamicsEnvWrapper(Env):
    """
    A wrapper that wrap the dynamics into a standard baconian env
    """

    @typechecked
    def __init__(self, dynamics: DynamicsModel, name: str = 'dynamics_env'):
        super().__init__(name)
        self._dynamics = dynamics
        self._reward_func = None
        self._terminal_func = None
        self.env_spec = dynamics.env_spec

    def step(self, action: np.ndarray, **kwargs):
        super().step(action)
        state = self.get_state() if 'state' not in kwargs else kwargs['state']
        new_state = self._dynamics.step(action=action, **kwargs)
        re = self._reward_func(state=state, new_state=new_state, action=action)
        terminal = self._terminal_func(state=state, action=action, new_state=new_state)
        return new_state, re, terminal, ()

    def reset(self):
        super(DynamicsEnvWrapper, self).reset()
        self._dynamics.reset_state()
        return self.get_state()

    def init(self):
        super().init()
        self._dynamics.init()

    def get_state(self):
        return self._dynamics.state

    def seed(self, seed=None):
        ConsoleLogger().print('warning', 'seed on dynamics model has no effect ')
        pass

    def save(self, *args, **kwargs):
        return self._dynamics.save(*args, **kwargs)

    def load(self, *args, **kwargs):
        return self._dynamics.load(*args, **kwargs)

    def set_terminal_reward_func(self, terminal_func: TerminalFunc, reward_func: RewardFunc):
        self._terminal_func = terminal_func
        self._reward_func = reward_func


class DynamicsPriorModel(Basic):
    def __init__(self, env_spec: EnvSpec, parameters: Parameters, name: str):
        super().__init__(name=name)
        self.env_spec = env_spec
        self.parameters = parameters



import os
from baconian.core.util import init_func_arg_record_decorator
import baconian.common.files as files


class Config(object):
    pass


class DictConfig(Config):

    @init_func_arg_record_decorator()
    def __init__(self, required_key_dict: dict, config_dict: dict = None, cls_name=""):
        self.cls_name = cls_name

        self.required_key_dict = required_key_dict
        if config_dict:
            self._config_dict = config_dict
        else:
            self._config_dict = {}

    @property
    def config_dict(self):
        return self._config_dict

    @config_dict.setter
    def config_dict(self, new_value):
        if self.check_config(dict=new_value, key_dict=self.required_key_dict) is True:
            for key, val in new_value.items():
                if type(val) is list:
                    new_value[str(key)] = tuple(val)
            self._config_dict = new_value
            for key, val in self._config_dict.items():
                setattr(self, key, val)

    def save_config(self, path, name):
        DictConfig.save_to_json(dict=self.config_dict, path=path, file_name=name)

    def load_config(self, path):
        res = DictConfig.load_json(file_path=path)
        self.config_dict = res

    def check_config(self, dict: dict, key_dict: dict) -> bool:
        if self.check_dict_key(check_dict=dict, required_key_dict=key_dict):
            return True
        else:
            return False

    @staticmethod
    def load_json(file_path):
        return files.load_json(file_path)

    @staticmethod
    def save_to_json(dict, path, file_name=None):
        if file_name is not None:
            path = os.path.join(path, file_name)
        files.save_to_json(dict, path=path, file_name=file_name)

    def check_dict_key(self, check_dict: dict, required_key_dict: dict) -> bool:
        for key, val in required_key_dict.items():
            if not isinstance(check_dict, dict):
                raise TypeError('{}: input check dict should be a dict instead of {}'.format(self.cls_name,
                                                                                             type(check_dict).__name__))
            if key not in check_dict:
                raise IndexError('{} Missing Key {}'.format(self.cls_name, key))
            if required_key_dict[key] is not None and not isinstance(check_dict[key], type(required_key_dict[key])):
                raise TypeError('{} should be type {} from required key dict file but with type {}'.
                                format(key, type(required_key_dict[key]), type(check_dict[key])))
            if isinstance(val, dict):
                self.check_dict_key(check_dict=check_dict[key], required_key_dict=required_key_dict[key])
        return True

    def __call__(self, key):
        if key not in self.config_dict:
            raise KeyError('{} key {} not in the config'.format(self.cls_name, key))
        else:
            return self.config_dict[key]

    def __getitem__(self, item):
        return self.__call__(item)

    def set(self, key, val):
        self.config_dict[key] = val


import unittest
from baconian.core.core import EnvSpec
from baconian.common.spaces import *
from baconian.common.special import *
from baconian.envs.gym_env import make
import numpy as np
from baconian.test.tests.set_up.setup import *


class TestSpace(TestWithLogSet):
    def test_box(self):
        pass



from baconian.test.tests.set_up.setup import TestTensorflowSetup
from baconian.core.util import get_global_arg_dict


class TestCopyGlobally(TestTensorflowSetup):
    def test_init_arg_decorator(self):
        dqn, local = self.create_dqn()
        env_spec = local['env_spec']
        mlp_q = local['mlp_q']
        dqn.init()
        a = get_global_arg_dict()
        self.assertTrue(dqn in a)
        self.assertTrue(env_spec in a)
        self.assertTrue(mlp_q in a)
        print(a.keys())
        self.setUp()

        a = get_global_arg_dict()



"""
From openai baselines
"""
import numpy as np
from typeguard import typechecked
from baconian.common.schedules import Scheduler


class AdaptiveParamNoiseSpec(object):
    def __init__(self, initial_stddev=0.1, desired_action_stddev=0.1, adoption_coefficient=1.01):
        self.initial_stddev = initial_stddev
        self.desired_action_stddev = desired_action_stddev
        self.adoption_coefficient = adoption_coefficient

        self.current_stddev = initial_stddev

    def adapt(self, distance):
        if distance > self.desired_action_stddev:
            # Decrease stddev.
            self.current_stddev /= self.adoption_coefficient
        else:
            # Increase stddev.
            self.current_stddev *= self.adoption_coefficient

    def get_stats(self):
        stats = {
            'param_noise_stddev': self.current_stddev,
        }
        return stats

    def __repr__(self):
        fmt = 'AdaptiveParamNoiseSpec(initial_stddev={}, desired_action_stddev={}, adoption_coefficient={})'
        return fmt.format(self.initial_stddev, self.desired_action_stddev, self.adoption_coefficient)


class ActionNoise(object):
    def reset(self):
        pass

    def __call__(self, *args, **kwargs):
        return 0.0


class NormalActionNoise(ActionNoise):
    def __init__(self, mu=0.0, sigma=1.0):
        self.mu = mu
        self.sigma = sigma

    def __call__(self):
        return np.random.normal(self.mu, self.sigma)

    def __repr__(self):
        return 'NormalActionNoise(mu={}, sigma={})'.format(self.mu, self.sigma)


# Based on http://math.stackexchange.com/questions/1287634/implementing-ornstein-uhlenbeck-in-matlab
class OrnsteinUhlenbeckActionNoise(ActionNoise):
    def __init__(self, mu, sigma, theta=.15, dt=1e-2, x0=None):
        self.theta = theta
        self.mu = mu
        self.sigma = sigma
        self.dt = dt
        self.x0 = x0
        self.reset()

    def __call__(self):
        x = self.x_prev + self.theta * (self.mu - self.x_prev) * self.dt + self.sigma * np.sqrt(
            self.dt) * np.random.normal(size=self.mu.shape)
        self.x_prev = x
        return x

    def reset(self):
        self.x_prev = self.x0 if self.x0 is not None else np.zeros_like(self.mu)

    def __repr__(self):
        return 'OrnsteinUhlenbeckActionNoise(mu={}, sigma={})'.format(self.mu, self.sigma)


class UniformNoise(ActionNoise):
    def __init__(self, scale):
        self.scale = scale

    def __call__(self):
        uniformNoiseValue = self.scale * (np.random.rand() - 0.5)
        return uniformNoiseValue


class OUNoise(ActionNoise):
    def __init__(self, theta=0.05, sigma=0.25, init_state=0.0):
        self.theta = theta
        self.sigma = sigma
        self.state = init_state

    def __call__(self):
        state = self.state - self.theta * self.state + self.sigma * np.random.randn()
        self.state = state
        return self.state

    def reset(self):
        self.state = 0.0


class AgentActionNoiseWrapper(object):
    INJECT_TYPE = ['WEIGHTED_DECAY', '']

    @typechecked
    def __init__(self, noise: ActionNoise, action_weight_scheduler: Scheduler, noise_weight_scheduler: Scheduler):
        self.action_weight_scheduler = action_weight_scheduler
        self.noise_weight_scheduler = noise_weight_scheduler
        self.noise = noise

    def __call__(self, action, **kwargs):
        noise = self.noise()
        return self.action_weight_scheduler.value() * action + self.noise_weight_scheduler.value() * noise

    def reset(self):
        self.noise.reset()



import unittest
from baconian.common.special import *
from baconian.test.tests.set_up.setup import TestTensorflowSetup


class TestMLPQValueFunction(TestTensorflowSetup):

    def test_init(self):
        mlp_q, local = self.create_mlp_q_func(name='mlp_q')
        env = local['env']
        env_spec = local['env_spec']
        mlp_q.init()

        action = env.action_space.sample()
        action = np.array([action])
        action = flatten_n(env_spec.action_space, action)
        mlp_q.forward(obs=env.observation_space.sample(), action=action)

        action = env.action_space.sample()
        action = np.array([action])
        action = flatten_n(space=env.action_space, obs=action)

        mlp_q.forward(obs=env.observation_space.sample(), action=action)

        action = env.action_space.sample()
        action = np.array([action])
        action = flatten_n(space=mlp_q.env_spec.action_space,
                           obs=make_batch(action, original_shape=mlp_q.env_spec.action_shape))

        mlp_q.forward(obs=env.observation_space.sample(), action=action)

    def test_copy(self):
        mlp_q, local = self.create_mlp_q_func(name='mlp_q')
        mlp_q.init()

        new_mlp = mlp_q.make_copy(name='new_mlp',
                                  name_scope='mlp_q',
                                  reuse=True)

        new_mlp.init()

        self.assertGreater(len(mlp_q.parameters('tf_var_list')), 0)
        self.assertGreater(len(new_mlp.parameters('tf_var_list')), 0)

        for var1, var2 in zip(mlp_q.parameters('tf_var_list'), new_mlp.parameters('tf_var_list')):
            self.assertEqual(var1.shape, var2.shape)
            self.assertEqual(id(var1), id(var2))

        not_reuse_mlp = mlp_q.make_copy(name='no-reuse-mlp',
                                        name_scope='mlp_no_reuse',
                                        reuse=False)
        not_reuse_mlp.init()
        self.assertGreater(len(not_reuse_mlp.parameters('tf_var_list')), 0)

        for var1, var2 in zip(mlp_q.parameters('tf_var_list'), not_reuse_mlp.parameters('tf_var_list')):
            self.assertEqual(var1.shape, var2.shape)
            self.assertNotEqual(id(var1), id(var2))


if __name__ == '__main__':
    unittest.main()



"""
Calculate the max sum reward for mountaincar experiment based across multiple trials
"""

import json

# Declare number of trials here (Example: if numTrials is 10, it will search through trial 1-10)
numTrials = 10
maxRewardArray = []

for i in range(numTrials):
    with open('./mountain_log_path' + str(i+1) + '/record/benchmark_agent/TEST/log.json', 'r') as f:
        result_dict = json.load(f)

    x = []
    for result in result_dict["sum_reward"]:
        x.append(result["log_val"]["__ndarray__"])

    print("TESTSET: " + str(i+1) + " - MAX:", max(x))
    maxRewardArray.append(max(x))

print("-----AVERAGE OF 10 Test Sets: ", max(maxRewardArray), "-------")






import unittest
from baconian.core.core import EnvSpec
from baconian.common.spaces import *
from baconian.common.special import *
from baconian.envs.gym_env import make
import numpy as np
from baconian.test.tests.set_up.setup import BaseTestCase


def create_env_spec():
    env = EnvSpec(obs_space=Box(low=0.0,
                                high=1.0,
                                shape=[2]),
                  action_space=Box(low=0.0,
                                   high=1.0,
                                   shape=[2]))
    return env


class TestMisc(BaseTestCase):

    def test_env_spec(self):
        env = create_env_spec()
        self.assertEqual(env.flat_obs_dim, 2)
        self.assertEqual(env.flat_action_dim, 2)

    def test_misc_func(self):
        env = make('Pendulum-v0')
        a = make_batch(v=np.array([env.action_space.sample() for _ in range(10)]),
                       original_shape=env.env_spec.action_shape)
        self.assertEqual(a.shape[0], 10)
        self.assertTrue(a.shape[1:] == env.env_spec.action_shape)
        for ac in a:
            self.assertTrue(env.action_space.contains(ac))

        a = make_batch(v=np.array([env.observation_space.sample() for _ in range(10)]),
                       original_shape=env.env_spec.obs_shape)
        self.assertTrue(a.shape[1:] == env.env_spec.obs_shape)
        for ac in a:
            self.assertTrue(env.observation_space.contains(ac))

        env = make('Acrobot-v1')
        a = make_batch(v=np.array([env.action_space.sample() for _ in range(10)]),
                       original_shape=env.env_spec.action_shape)
        self.assertEqual(a.shape[0], 10)

        self.assertTrue(a.shape[1:] == env.env_spec.action_shape)
        for ac in a:
            self.assertTrue(env.action_space.contains(ac))

        a = make_batch(v=np.array([env.observation_space.sample() for _ in range(10)]),
                       original_shape=env.env_spec.obs_shape)
        self.assertEqual(a.shape[0], 10)
        self.assertTrue(a.shape[1:] == env.env_spec.obs_shape)
        for ac in a:
            self.assertTrue(env.observation_space.contains(ac))



from baconian.core.core import Basic
import abc
import numpy as np


class TerminalFunc(Basic):
    allow_duplicate_name = True

    def __init__(self, name='terminal_func'):
        super().__init__(name=name)

    @abc.abstractmethod
    def __call__(self, state, action, new_state, **kwargs) -> bool:
        raise NotImplementedError

    def init(self):
        pass


class RandomTerminalFunc(TerminalFunc):
    """
    Debug and test use only
    """

    def __init__(self, name='random_terminal_func'):
        super().__init__(name)

    def __call__(self, state=None, action=None, new_state=None, **kwargs) -> bool:
        return np.random.random() > 0.5


class FixedEpisodeLengthTerminalFunc(Basic):

    def __init__(self, max_step_length: int, step_count_fn, status=None,
                 name: str = 'fixed_epsiode_length_terminal_func'):
        super().__init__(name, status)
        self.max_step_length = max_step_length
        self.step_count_fn = step_count_fn

    def __call__(self, state=None, action=None, new_state=None, **kwargs) -> bool:
        if self.step_count_fn() >= self.max_step_length:
            return True
        else:
            return False



from baconian.benchmark.ddpg_benchmark import mountiancar_task_fn, pendulum_task_fn
from baconian.benchmark.dyna_benchmark import dyna_pendulum_task_fn
from baconian.benchmark.mpc_benchmark import mpc_pendulum_task_fn
from baconian.benchmark.ppo_benchmark import pendulum_task_fn as ppo_pendulum_task_fn
from baconian.benchmark.ppo_benchmark import reacher_task_fn
from baconian.benchmark.ppo_benchmark import swimmer_task_fn
from baconian.benchmark.ppo_benchmark import hopper_task_fn
from baconian.benchmark.ppo_benchmark import half_cheetah_task_fn
from baconian.benchmark.ppo_benchmark import inverted_pendulum_task_fn, half_cheetah_bullet_env_task_fn
from baconian.benchmark.iLQR_benchmark import ilqr_pendulum_task_fn
from baconian.benchmark.dqn_benchmark import acrobot_task_fn, lunarlander_task_fn
import argparse
import os
import time
from baconian.config.global_config import GlobalConfig
from baconian.core.experiment_runner import duplicate_exp_runner

arg = argparse.ArgumentParser()
env_id_to_task_fn = {
    'Pendulum-v0': {
        'ddpg': pendulum_task_fn,
        'dyna': dyna_pendulum_task_fn,
        'mpc': mpc_pendulum_task_fn,
        'ppo': ppo_pendulum_task_fn,
        'ilqr': ilqr_pendulum_task_fn
    },
    'MountainCarContinuous-v0': {
        'ddpg': mountiancar_task_fn,
    },
    'InvertedPendulum-v2': {
        'ppo': inverted_pendulum_task_fn,
    },
    'Reacher-v2': {
        'ppo': reacher_task_fn,
    },
    'Swimmer-v2': {
        'ppo': swimmer_task_fn,
    },
    'Hopper-v2': {
        'ppo': hopper_task_fn,
    },
    'HalfCheetah-v2': {
        'ppo': half_cheetah_task_fn,
    },
    'HalfCheetahBulletEnv-v0': {
        'ppo': half_cheetah_bullet_env_task_fn,

    },
    'Acrobot-v1': {
        'dqn': acrobot_task_fn,
    },
    'LunarLander-v2': {
        'dqn': lunarlander_task_fn,
    }
}
alog_list = ['ddpg', 'dyna', 'mpc', 'ppo', 'ilqr', 'dqn']

arg.add_argument('--env_id', type=str, choices=list(env_id_to_task_fn.keys()))
arg.add_argument('--algo', type=str, choices=alog_list)
arg.add_argument('--count', type=int, default=1)
arg.add_argument('--cuda_id', type=int, default=-1)
args = arg.parse_args()


if __name__ == '__main__':
    CURRENT_PATH = os.path.dirname(os.path.realpath(__file__))

    GlobalConfig().set('DEFAULT_LOG_PATH', os.path.join(CURRENT_PATH, 'benchmark_log', args.env_id, args.algo,
                                                        time.strftime("%Y-%m-%d_%H-%M-%S")))
    ExpRootPath = GlobalConfig().DEFAULT_LOG_PATH
    duplicate_exp_runner(args.count, env_id_to_task_fn[args.env_id][args.algo], gpu_id=args.cuda_id)



from baconian.core.core import Env, EnvSpec
import gym.envs
from gym.envs.registration import registry
import roboschool
import pybullet
import pybullet_envs

have_mujoco_flag = True
try:
    from gym.envs.mujoco import mujoco_env
except Exception:
    have_mujoco_flag = False
import numpy as np
import types
import gym.spaces as GymSpace
import baconian.common.spaces as garage_space
import gym.error as gym_error

_env_inited_count = dict()


def make(gym_env_id: str, allow_multiple_env=True):
    """

    :param gym_env_id: gym environment id
    :type gym_env_id: int
    :param allow_multiple_env: allow multiple environments, by default True
    :type allow_multiple_env: bool
    :return: new gym environment
    :rtype: GymEnv
    """
    if allow_multiple_env is True:

        if gym_env_id not in _env_inited_count:
            _env_inited_count[gym_env_id] = 0
        else:
            _env_inited_count[gym_env_id] += 1

        return GymEnv(gym_env_id, name='{}_{}'.format(gym_env_id, _env_inited_count[gym_env_id]))
    else:
        return GymEnv(gym_env_id)


def space_converter(space: GymSpace.Space):
    """
    Convert space into any one of "Box", "Discrete", or "Tuple" type.

    :param space: space of gym environment
    :type space: GymSpace
    :return: converted space
    :rtype: Box, Discrete, or Tuple
    """
    if isinstance(space, GymSpace.Box):
        return garage_space.Box(low=space.low, high=space.high)
    elif isinstance(space, GymSpace.Dict):
        return garage_space.Dict(space.spaces)
    elif isinstance(space, GymSpace.Discrete):
        return garage_space.Discrete(space.n)
    elif isinstance(space, GymSpace.Tuple):
        return garage_space.Tuple(list(map(space_converter, space.spaces)))
    else:
        raise NotImplementedError


class GymEnv(Env):
    """
    Gym environment wrapping module
    """
    _all_gym_env_id = list(registry.env_specs.keys())

    def __init__(self, gym_env_id: str, name: str = None):
        """

        :param gym_env_id: gym environment id
        :type gym_env_id: str
        :param name: name of the gym environment instance
        :type name: str
        """
        super().__init__(name=name if name else gym_env_id)
        self.env_id = gym_env_id
        try:
            self._gym_env = gym.make(gym_env_id)
        except gym_error.UnregisteredEnv:
            raise ValueError('Env id: {} is not supported currently'.format(gym_env_id))
        self._gym_env = gym.make(gym_env_id)
        self.action_space = space_converter(self._gym_env.action_space)
        self.observation_space = space_converter(self._gym_env.observation_space)
        if isinstance(self.action_space, garage_space.Box):
            self.action_space.low = np.nan_to_num(self.action_space.low)
            self.action_space.high = np.nan_to_num(self.action_space.high)
            self.action_space.sample = types.MethodType(self._sample_with_nan, self.action_space)
        if isinstance(self.observation_space, garage_space.Box):
            self.observation_space.low = np.nan_to_num(self.observation_space.low)
            self.observation_space.high = np.nan_to_num(self.observation_space.high)
            self.observation_space.sample = types.MethodType(self._sample_with_nan, self.observation_space)
        self.env_spec = EnvSpec(obs_space=self.observation_space,
                                action_space=self.action_space)

        self.reward_range = self._gym_env.reward_range

    def step(self, action):
        """

        :param action: action to be taken by agent in the environment
        :type action: action to be taken by agent in the environment
        :return: step of the unwrapped environment
        :rtype: gym env
        """
        super().step(action)
        action = self.env_spec.flat_action(action)
        return self.unwrapped.step(action=action)

    def reset(self):
        """
        Reset the gym environment.
        :return:
        """
        super().reset()

        return self.unwrapped.reset()

    def init(self):
        """
        Initialize the gym environment.
        :return:
        """
        super().init()

        return self.reset()

    def seed(self, seed=None):
        """

        :param seed: seed of random number generalization
        :type seed: int
        :return: seed of the unwrapped environment
        :rtype: int
        """
        return super().seed(seed)

    def get_state(self):
        """

        :return:the state of unwrapped gym environment
        :rtype: np.ndarray
        """
        if (have_mujoco_flag is True and isinstance(self.unwrapped_gym, mujoco_env.MujocoEnv)) or (
                hasattr(self.unwrapped_gym, '_get_obs') and callable(self.unwrapped_gym._get_obs)):
            return self.unwrapped_gym._get_obs()
        elif hasattr(self.unwrapped_gym, '_get_ob') and callable(self.unwrapped_gym._get_ob):
            return self.unwrapped_gym._get_ob()
        elif hasattr(self.unwrapped_gym, 'state'):
            return self.unwrapped_gym.state if isinstance(self.unwrapped_gym.state, np.ndarray) else np.array(
                self.unwrapped_gym.state)
        elif hasattr(self.unwrapped_gym, 'observation'):
            return self.unwrapped_gym.observation if isinstance(self.unwrapped_gym.observation,
                                                                np.ndarray) else np.array(
                self.unwrapped_gym.state)
        elif hasattr(self.unwrapped_gym, 'spec') and hasattr(self.unwrapped_gym.spec,
                                                             'id') and self.unwrapped_gym.spec.id in specialEnv:
            return specialEnv[self.unwrapped_gym.spec.id](self)
        elif hasattr(self.unwrapped_gym, 'robot'):
            return self.unwrapped_gym.robot.calc_state()
        else:
            raise ValueError('Env id: {} is not supported for method get_state'.format(self.env_id))

    @property
    def unwrapped(self):
        """
        :return: original unwrapped gym environment
        :rtype: gym env
        """
        return self._gym_env

    @property
    def unwrapped_gym(self):
        """

        :return: gym environment, depend on attribute 'unwrapped'
        :rtype: gym env
        """
        if hasattr(self._gym_env, 'unwrapped'):
            return self._gym_env.unwrapped
        else:
            return self._gym_env

    @staticmethod
    def _sample_with_nan(space: garage_space.Space):
        """

        :param space: a 'Box'type space
        :return: numpy clip of space that contains nan values
        :rtype: np.ndarray
        """
        assert isinstance(space, garage_space.Box)
        high = np.ones_like(space.low)
        low = -1 * np.ones_like(space.high)
        return np.clip(np.random.uniform(low=low, high=high, size=space.low.shape),
                       a_min=space.low,
                       a_max=space.high)

    def __str__(self):
        return "<GymEnv instance> {}".format(self.env_id)


def get_lunarlander_state(env):
    pos = env.unwrapped_gym.lander.position
    vel = env.unwrapped_gym.lander.linearVelocity
    fps = 50
    scale = 30.0  # affects how fast-paced the game is, forces should be adjusted as well
    leg_down = 18
    viewport_w = 600
    viewport_h = 400
    state = [
        (pos.x - viewport_w / scale / 2) / (viewport_w / scale / 2),
        (pos.y - (env.unwrapped_gym.helipad_y + leg_down / scale)) / (viewport_h / scale / 2),
        vel.x * (viewport_w / scale / 2) / fps,
        vel.y * (viewport_h / scale / 2) / fps,
        env.unwrapped_gym.lander.angle,
        20.0 * env.unwrapped_gym.lander.angularVelocity / fps,
        1.0 if env.unwrapped_gym.legs[0].ground_contact else 0.0,
        1.0 if env.unwrapped_gym.legs[1].ground_contact else 0.0
    ]
    return np.array(state, dtype=np.float32)


specialEnv = {
    'LunarLander-v2': get_lunarlander_state
}



from baconian.envs.gym_env import GymEnv
from baconian.test.tests.set_up.setup import TestWithLogSet
from gym import make


class TestEnv(TestWithLogSet):
    def test_gym_env(self):
        a = GymEnv('RoboschoolHalfCheetah-v1')
        a.set_status('TRAIN')
        self.assertEqual(a.total_step_count_fn(), 0)
        self.assertEqual(a._last_reset_point, 0)
        a.init()
        a.seed(10)
        a.reset()
        self.assertEqual(a.total_step_count_fn(), 0)
        self.assertEqual(a._last_reset_point, 0)
        for i in range(1000):
            new_st, re, done, _ = a.step(action=a.action_space.sample())
            self.assertEqual(a.total_step_count_fn(), i + 1)
            if done is True:
                a.reset()
                self.assertEqual(a._last_reset_point, a.total_step_count_fn())
                self.assertEqual(a._last_reset_point, i + 1)



def _get_copy_arg_with_tf_reuse(obj, kwargs: dict):
    # kwargs = deepcopy(kwargs)
    if 'reuse' in kwargs:
        if kwargs['reuse'] is True:
            if 'name_scope' in kwargs and kwargs['name_scope'] != obj.name_scope:
                raise ValueError('If reuse, the name scope should be same instead of : {} and {}'.format(
                    kwargs['name_scope'], obj.name_scope))
            else:
                kwargs.update(name_scope=obj.name_scope)
        else:
            if 'name_scope' in kwargs and kwargs['name_scope'] == obj.name:
                raise ValueError(
                    'If not reuse, the name scope should be different instead of: {} and {}'.format(
                        kwargs['name_scope'], obj.name_scope))
    return kwargs



from setuptools import setup, find_packages
import os


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]


CURRENT_PATH = os.path.dirname(os.path.realpath(__file__))

req = parse_requirements(filename=os.path.join(CURRENT_PATH, 'requirements.txt'))

# req = [str(ir.req) for ir in req]

# print(req)
with open(os.path.join(CURRENT_PATH, 'README.md'), 'r', encoding='utf-8') as f:
    long_description = f.read()
exec(open('baconian/version.py').read())

ver = __version__

setup(
    name='baconian',
    version=ver,
    url='https://github.com/cap-ntu/baconian-project',
    license='MIT License',
    author='Linsen Dong',
    author_email='linsen001@e.ntu.edu.sg',
    description='model-based reinforcement learning toolbox',
    install_requires=req,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(include=['baconian', 'baconian.*'], exclude=[]),
    python_requires='>=3.5',
    include_package_data=True,
    package_data={'baconian': ['config/required_keys/*', 'benchmark/**/*', './*']}
)



"""
This script is from garage
"""
import gym.spaces
import numpy as np
import scipy
import scipy.signal
from typeguard import typechecked
import baconian.common.spaces as mbrl_spaces


def weighted_sample(weights, objects):
    """
    Return a random item from objects, with the weighting defined by weights
    (which must sum to 1).
    """
    # An array of the weights, cumulatively summed.
    cs = np.cumsum(weights)
    # Find the index of the first weight over a random value.
    idx = sum(cs < np.random.rand())
    return objects[min(idx, len(objects) - 1)]


def weighted_sample_n(prob_matrix, items):
    s = prob_matrix.cumsum(axis=1)
    r = np.random.rand(prob_matrix.shape[0])
    k = (s < r.reshape((-1, 1))).sum(axis=1)
    n_items = len(items)
    return items[np.minimum(k, n_items - 1)]


# compute softmax for each row
def softmax(x):
    shifted = x - np.max(x, axis=-1, keepdims=True)
    expx = np.exp(shifted)
    return expx / np.sum(expx, axis=-1, keepdims=True)


# compute entropy for each row
def cat_entropy(x):
    return -np.sum(x * np.log(x), axis=-1)


# compute perplexity for each row
def cat_perplexity(x):
    return np.exp(cat_entropy(x))


def explained_variance_1d(ypred, y):
    assert y.ndim == 1 and ypred.ndim == 1
    vary = np.var(y)
    if np.isclose(vary, 0):
        if np.var(ypred) > 0:
            return 0
        else:
            return 1
    return 1 - np.var(y - ypred) / (vary + 1e-8)


def to_onehot(ind, dim):
    ret = np.zeros(dim)
    ret[ind] = 1
    return ret


def to_onehot_n(inds, dim):
    ret = np.zeros((len(inds), dim))
    ret[np.arange(len(inds)), inds] = 1
    return ret


def from_onehot(v):
    return np.nonzero(v)[0][0]


def from_onehot_n(v):
    if ((isinstance(v, np.ndarray) and not v.size)
            or (isinstance(v, list) and not v)):
        return []
    return np.nonzero(v)[1]


def discount_cumsum(x, discount):
    # See https://docs.scipy.org/doc/scipy/reference/tutorial/signal.html#difference-equation-filtering  # noqa: E501
    # Here, we have y[t] - discount*y[t+1] = x[t]
    # or rev(y)[t] - discount*rev(y)[t-1] = rev(x)[t]
    return scipy.signal.lfilter(
        [1], [1, float(-discount)], x[::-1], axis=0)[::-1]


def discount_return(x, discount):
    return np.sum(x * (discount ** np.arange(len(x))))


def rk4(derivs, y0, t, *args, **kwargs):
    """
    Integrate 1D or ND system of ODEs using 4-th order Runge-Kutta.
    This is a toy implementation which may be useful if you find
    yourself stranded on a system w/o scipy.  Otherwise use
    :func:`scipy.integrate`.

    *y0*
        initial state vector

    *t*
        sample times

    *derivs*
        returns the derivative of the system and has the
        signature ``dy = derivs(yi, ti)``

    *args*
        additional arguments passed to the derivative function

    *kwargs*
        additional keyword arguments passed to the derivative function

    Example 1 ::

        ## 2D system

        def derivs6(x,t):
            d1 =  x[0] + 2*x[1]
            d2 =  -3*x[0] + 4*x[1]
            return (d1, d2)
        dt = 0.0005
        t = arange(0.0, 2.0, dt)
        y0 = (1,2)
        yout = rk4(derivs6, y0, t)

    Example 2::

        ## 1D system
        alpha = 2
        def derivs(x,t):
            return -alpha*x + exp(-t)

        y0 = 1
        yout = rk4(derivs, y0, t)


    If you have access to scipy, you should probably be using the
    scipy.integrate tools rather than this function.
    """

    try:
        ny = len(y0)
    except TypeError:
        yout = np.zeros((len(t),), np.float_)
    else:
        yout = np.zeros((len(t), ny), np.float_)

    yout[0] = y0
    i = 0

    for i in np.arange(len(t) - 1):
        thist = t[i]
        dt = t[i + 1] - thist
        dt2 = dt / 2.0
        y0 = yout[i]

        k1 = np.asarray(derivs(y0, thist, *args, **kwargs))
        k2 = np.asarray(derivs(y0 + dt2 * k1, thist + dt2, *args, **kwargs))
        k3 = np.asarray(derivs(y0 + dt2 * k2, thist + dt2, *args, **kwargs))
        k4 = np.asarray(derivs(y0 + dt * k3, thist + dt, *args, **kwargs))
        yout[i + 1] = y0 + dt / 6.0 * (k1 + 2 * k2 + 2 * k3 + k4)
    return yout


def make_batch(v, original_shape: (list, tuple)):
    if not isinstance(v, np.ndarray):
        v = np.array(v)
    # assert len(v.shape) <= len(original_shape) + 1
    if len(v.shape) == len(original_shape) + 1 and np.equal(np.array(v.shape[1:]),
                                                            np.array(original_shape)).all() is True:
        return v
    else:
        bs = np.prod(list(v.shape)) / np.prod(original_shape)
        assert float(bs).is_integer()
        return np.reshape(v, newshape=[int(bs)] + list(original_shape))


def flat_dim(space):
    if isinstance(space, mbrl_spaces.Box):
        return np.prod(space.low.shape)
    elif isinstance(space, mbrl_spaces.Discrete):
        return space.n
    elif isinstance(space, mbrl_spaces.Tuple):
        return np.sum([flat_dim(x) for x in space.spaces])
    else:
        raise NotImplementedError


def flatten(space, obs, one_hot_for_discrete=False):
    if isinstance(space, mbrl_spaces.Box):
        return np.asarray(obs).flatten()
    elif isinstance(space, mbrl_spaces.Discrete):
        if one_hot_for_discrete is True:
            if space.n == 2:
                obs = int(obs)
            return to_onehot(obs, space.n)
        else:
            return int(obs)
    elif isinstance(space, mbrl_spaces.Tuple):
        return np.concatenate(
            [flatten(c, xi) for c, xi in zip(space.spaces, obs)])
    else:
        raise NotImplementedError


def flatten_n(space, obs):
    if isinstance(space, mbrl_spaces.Box):
        obs = np.asarray(obs)
        return obs.reshape((obs.shape[0], -1))
    elif isinstance(space, mbrl_spaces.Discrete):
        return to_onehot_n(np.array(obs, dtype=np.int), space.n)
    elif isinstance(space, mbrl_spaces.Tuple):
        obs_regrouped = [[obs[i] for o in obs] for i in range(len(obs[0]))]
        flat_regrouped = [
            flatten_n(c, oi) for c, oi in zip(space.spaces, obs_regrouped)
        ]
        return np.concatenate(flat_regrouped, axis=-1)
    else:
        raise NotImplementedError


def unflatten(space, obs):
    if isinstance(space, mbrl_spaces.Box):
        return np.asarray(obs).reshape(space.shape)
    elif isinstance(space, mbrl_spaces.Discrete):
        return from_onehot(np.array(obs, dtype=np.int))
    elif isinstance(space, mbrl_spaces.Tuple):
        dims = [flat_dim(c) for c in space.spaces]
        flat_xs = np.split(obs, np.cumsum(dims)[:-1])
        return tuple(unflatten(c, xi) for c, xi in zip(space.spaces, flat_xs))
    else:
        raise NotImplementedError


def unflatten_n(space, obs):
    if isinstance(space, mbrl_spaces.Box):
        obs = np.asarray(obs)
        return obs.reshape((obs.shape[0],) + space.shape)
    elif isinstance(space, mbrl_spaces.Discrete):
        return from_onehot_n(np.array(obs, dtype=np.int))
    elif isinstance(space, mbrl_spaces.Tuple):
        dims = [flat_dim(c) for c in space.spaces]
        flat_xs = np.split(obs, np.cumsum(dims)[:-1], axis=-1)
        unflat_xs = [
            unflatten_n(c, xi) for c, xi in zip(space.spaces, flat_xs)
        ]
        unflat_xs_grouped = list(zip(*unflat_xs))
        return unflat_xs_grouped
    else:
        raise NotImplementedError



from baconian.core.core import Basic, EnvSpec
import typeguard as tg
from baconian.core.parameters import Parameters
import abc


class ValueFunction(Basic):

    @tg.typechecked
    def __init__(self, env_spec: EnvSpec, parameters: Parameters = None, name='value_func'):
        super().__init__(name)
        self.env_spec = env_spec
        self.parameters = parameters

    @property
    def obs_space(self):
        return self.env_spec.obs_space

    @property
    def action_space(self):
        return self.env_spec.action_space

    @tg.typechecked
    def copy_from(self, obj) -> bool:
        if not isinstance(obj, type(self)):
            raise TypeError('Wrong type of obj %s to be copied, which should be %s' % (type(obj), type(self)))
        return True

    @abc.abstractmethod
    def forward(self, *args, **kwargs):
        raise NotImplementedError

    @tg.typechecked
    def update(self, param_update_args: dict, *args, **kwargs):
        self.parameters.update(**param_update_args)

    @abc.abstractmethod
    def init(self, source_obj=None):
        raise NotImplementedError

    @abc.abstractmethod
    def make_copy(self, *args, **kwargs):
        raise NotImplementedError


class VValueFunction(ValueFunction):
    def __init__(self, state_input, env_spec: EnvSpec, parameters: Parameters = None, name='value_func'):
        self.state_input = state_input
        super().__init__(env_spec, parameters, name)


class QValueFunction(ValueFunction):
    def __init__(self, state_input, action_input, env_spec: EnvSpec, parameters: Parameters = None, name='value_func'):
        self.state_input = state_input
        self.action_input = action_input
        super().__init__(env_spec, parameters, name)



from baconian.common.sampler.sample_data import TransitionData, TrajectoryData
from baconian.envs.gym_env import make
from baconian.core.core import EnvSpec
import numpy as np
from baconian.test.tests.set_up.setup import BaseTestCase
from baconian.common.error import *


class TestSampleData(BaseTestCase):
    def test_transition_data(self):
        env = make('Acrobot-v1')
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)
        a = TransitionData(env_spec)
        st = env.reset()
        for i in range(100):
            ac = env_spec.action_space.sample()
            st_new, re, done, _ = env.step(action=ac)
            a.append(state=st, new_state=st_new, action=ac, done=done, reward=re)
        self.assertEqual(a.reward_set.shape[0], 100)
        self.assertEqual(a.done_set.shape[0], 100)
        self.assertEqual(a.action_set.shape[0], 100)
        self.assertEqual(a.state_set.shape[0], 100)
        self.assertEqual(a.new_state_set.shape[0], 100)

        self.assertEqual(a('reward_set').shape[0], 100)
        self.assertEqual(a('done_set').shape[0], 100)

        self.assertEqual(a('state_set').shape[0], 100)
        self.assertEqual(a('new_state_set').shape[0], 100)
        self.assertEqual(a('action_set').shape[0], 100)
        iterator = a.return_generator()
        count = 0
        for st, new_st, ac, reward, terminal in iterator:
            count += 1
            self.assertTrue(env_spec.action_space.contains(ac))
            self.assertTrue(env_spec.obs_space.contains(st))
            self.assertTrue(env_spec.obs_space.contains(new_st))
            self.assertTrue(np.isscalar(reward))
            self.assertTrue(isinstance(bool(terminal), bool))
        self.assertEqual(count, 100)

        a = TransitionData(obs_shape=list(np.array(env_spec.obs_space.sample()).shape),
                           action_shape=list(np.array(env_spec.action_space.sample()).shape))
        st = env.reset()
        for i in range(100):
            ac = env_spec.action_space.sample()
            st_new, re, done, _ = env.step(action=ac)
            a.append(state=st, new_state=st_new, action=ac, done=done, reward=re)
        self.assertEqual(a.reward_set.shape[0], 100)
        self.assertEqual(a.done_set.shape[0], 100)

        self.assertEqual(a.action_set.shape[0], 100)
        self.assertEqual(a.state_set.shape[0], 100)
        self.assertEqual(a.new_state_set.shape[0], 100)

        self.assertEqual(a('reward_set').shape[0], 100)
        self.assertEqual(a('done_set').shape[0], 100)

        self.assertEqual(a('state_set').shape[0], 100)
        self.assertEqual(a('new_state_set').shape[0], 100)
        self.assertEqual(a('action_set').shape[0], 100)

        self.assertTrue(np.equal(a.get_mean_of('state_set'), a.apply_op('state_set', np.mean)).all())
        self.assertTrue(np.equal(a.get_sum_of('state_set'), a.apply_op('state_set', np.sum)).all())

        self.assertTrue(np.equal(a.get_sum_of('reward_set'), a.apply_op('reward_set', np.sum)).all())
        self.assertTrue(np.equal(a.get_sum_of('reward_set'), a.apply_op('reward_set', np.sum)).all())

        self.assertTrue(np.equal(a.get_sum_of('action_set'), a.apply_op('action_set', np.sum)).all())
        self.assertTrue(np.equal(a.get_sum_of('action_set'), a.apply_op('action_set', np.sum)).all())
        self.assertTrue(np.equal(a.apply_op('state_set', np.max, axis=-1), np.max(a('state_set'), axis=-1)).all())

        tmp_action = a('action_set').copy()
        a.apply_transformation(set_name='action_set', func=lambda x: x * 2, direct_apply=False)
        self.assertTrue(np.equal(tmp_action, a('action_set')).all())
        a.apply_transformation(set_name='action_set', func=lambda x: x * 2, direct_apply=True)
        self.assertTrue(np.equal(tmp_action * 2.0, a('action_set')).all())
        try:
            a.apply_transformation(set_name='action_set', func=lambda _: np.array([1, 2, 3]), direct_apply=True)
        except TransformationResultedToDifferentShapeError as e:
            pass
        else:
            raise TypeError

        a.apply_transformation(set_name='action_set', func=lambda x: x // 2, direct_apply=True)
        self.assertTrue(np.equal(tmp_action, a('action_set')).all())

        index = np.arange(len(a._internal_data_dict['state_set'][0])).tolist()
        b = a.get_copy()
        a.shuffle(index=list(index))
        for i in range(len(index)):
            for key in a._internal_data_dict.keys():
                self.assertTrue(np.equal(np.array(a._internal_data_dict[key][0][i]),
                                         np.array(b._internal_data_dict[key][0][i])).all())

        iterator = a.return_generator()
        count = 0
        for st, new_st, ac, reward, terminal in iterator:
            count += 1
            self.assertTrue(env_spec.action_space.contains(ac))
            self.assertTrue(env_spec.obs_space.contains(st))
            self.assertTrue(env_spec.obs_space.contains(new_st))
            self.assertTrue(np.isscalar(reward))
            self.assertTrue(isinstance(bool(terminal), bool))
        self.assertEqual(count, 100)
        count = 0
        iter = a.return_generator(batch_size=10)
        for st, new_st, ac, reward, terminal in iter:
            self.assertEqual(len(st), 10)
            self.assertEqual(len(new_st), 10)
            self.assertEqual(len(ac), 10)
            self.assertEqual(len(reward), 10)
            self.assertEqual(len(terminal), 10)
            count += 1
        self.assertEqual(count, 10)
        count = 0
        iter = a.return_generator(batch_size=10, infinite_run=True)
        for st, new_st, ac, reward, terminal in iter:
            self.assertEqual(len(st), 10)
            self.assertEqual(len(new_st), 10)
            self.assertEqual(len(ac), 10)
            self.assertEqual(len(reward), 10)
            self.assertEqual(len(terminal), 10)
            count += 1
            if count > 20:
                break
        self.assertGreater(count, 20)

        a.append_new_set(name='test', data_set=np.ones_like(a._internal_data_dict['state_set'][0]),
                         shape=a._internal_data_dict['state_set'][1])
        iter = a.return_generator(batch_size=10,
                                  assigned_keys=('state_set', 'new_state_set', 'action_set', 'reward_set', 'done_set', 'test'))
        count = 0
        for st, new_st, ac, reward, terminal, test in iter:
            self.assertEqual(len(test), 10)
            count += 1
        self.assertEqual(count, 10)

        a.reset()
        self.assertEqual(a.reward_set.shape[0], 0)
        self.assertEqual(a.done_set.shape[0], 0)

        self.assertEqual(a.action_set.shape[0], 0)
        self.assertEqual(a.state_set.shape[0], 0)
        self.assertEqual(a.new_state_set.shape[0], 0)

        self.assertEqual(a('reward_set').shape[0], 0)
        self.assertEqual(a('done_set').shape[0], 0)

        self.assertEqual(a('state_set').shape[0], 0)
        self.assertEqual(a('new_state_set').shape[0], 0)
        self.assertEqual(a('action_set').shape[0], 0)

    def test_trajectory_data(self):
        env = make('Acrobot-v1')
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)
        a = TrajectoryData(env_spec)
        tmp_traj = TransitionData(env_spec)
        st = env.reset()
        re_list = []
        st_list = []
        for i in range(100):
            ac = env_spec.action_space.sample()
            st_new, re, done, _ = env.step(action=ac)
            st_list.append(st_new)
            re_list.append(re)
            if (i + 1) % 10 == 0:
                done = True
            else:
                done = False
            tmp_traj.append(state=st, new_state=st_new, action=ac, done=done, reward=re)
            if done:
                a.append(tmp_traj.get_copy())
                tmp_traj.reset()
        self.assertEqual(a.trajectories.__len__(), 10)
        for traj in a.trajectories:
            self.assertEqual(len(traj), 10)
        data = a.return_as_transition_data()
        data_gen = data.return_generator()
        for d, re, st in zip(data_gen, re_list, st_list):
            self.assertEqual(d[3], re)
            self.assertTrue(np.equal(st, d[1]).all())



# import numpy as np
from scipy.linalg import inv

from baconian.algo.policy.policy import DeterministicPolicy
from baconian.core.parameters import Parameters
from baconian.core.core import EnvSpec
from baconian.algo.dynamics.linear_dynamics_model import LinearDynamicsModel
import autograd.numpy as np
from baconian.common.special import *
from baconian.algo.dynamics.reward_func.reward_func import CostFunc, QuadraticCostFunc


class LQR(object):

    def __init__(self, env_spec: EnvSpec, T, dyna_model: LinearDynamicsModel, cost_fn: QuadraticCostFunc):

        self.env_spec = env_spec
        self.T = T
        self.dyn_model = dyna_model
        self.cost_fn = cost_fn
        self.control_low = self.env_spec.action_space.low
        self.control_high = self.env_spec.action_space.high
        self.K, self.k, self.std = None, None, None
        self.C = np.repeat(np.expand_dims(self.cost_fn.C, axis=0), [self.T], axis=0)
        self.F = np.repeat(np.expand_dims(self.dyn_model.F, axis=0), [self.T], axis=0)
        self.c = np.repeat(np.expand_dims(self.cost_fn.c, axis=0), [self.T], axis=0)
        self.f = np.repeat(np.expand_dims(self.dyn_model.f, axis=0), [self.T], axis=0)

    def differentiate(self):

        "get gradient values using finite difference"

        C = np.repeat(np.expand_dims(self.cost_fn.C, axis=0), [self.T, 1, 1])
        F = np.repeat(np.expand_dims(self.dyn_model.F, axis=0), [self.T, 1, 1])
        c = np.repeat(np.expand_dims(self.cost_fn.c, axis=0), (self.T, 1))
        f = np.repeat(np.expand_dims(self.dyn_model.f, axis=0), (self.T, 1))
        return C, F, c, f

    def backward(self, x_seq, u_seq):

        "initialize F_t, C_t, f_t, c_t, V_t, v_t"
        # todo C, F, c, f can be analytical set here
        C, F, c, f = self.C, self.F, self.c, self.f

        n = x_seq[0].shape[0]

        "initialize V_t1 and v_t1"

        c_x = c[-1][:n]
        c_u = c[-1][n:]

        C_xx = C[-1][:n, :n]
        C_xu = C[-1][:n, n:]
        C_ux = C[-1][n:, :n]
        C_uu = C[-1][n:, n:]

        # C_uu = C_uu + self.mu * np.eye(C_uu.shape[0])

        K = np.zeros((self.T + 1, u_seq[0].shape[0], x_seq[0].shape[0]))
        k = np.zeros((self.T + 1, u_seq[0].shape[0]))

        V = np.zeros((self.T + 1, x_seq[0].shape[0], x_seq[0].shape[0]))
        v = np.zeros((self.T + 1, x_seq[0].shape[0]))

        # K[-1] = -np.dot(inv(C_uu), C_ux)
        # k[-1] = -np.dot(inv(C_uu), c_u)

        # V[-1] = C_xx + np.dot(C_xu, K[-1]) + np.dot(K[-1].T, C_ux) + np.dot(np.dot(K[-1].T, C_uu), K[-1])
        # v[-1] = c_x + np.dot(C_xu, k[-1]) + np.dot(K[-1].T, c_u) + np.dot(np.dot(K[-1].T, C_uu), k[-1])
        V[-1] = C_xx
        v[-1] = c_x

        "initialize Q_t1 and q_t1"

        Q = list(np.zeros((self.T)))
        q = list(np.zeros((self.T)))

        self.std = []
        "loop till horizon"

        t = self.T - 1
        while t >= 0:
            Q[t] = C[t] + np.dot(np.dot(F[t].T, V[t + 1]), F[t])
            q[t] = c[t] + np.dot(F[t].T, v[t + 1]) + np.dot(np.dot(F[t].T, V[t + 1]), f[t])

            "differentiate Q to get Q_uu, Q_xx, Q_ux, Q_u, Q_x"

            q_x = q[t][:n]
            q_u = q[t][n:]

            Q_xx = Q[t][:n, :n]
            Q_xu = Q[t][:n, n:]
            Q_ux = Q[t][n:, :n]
            Q_uu = Q[t][n:, n:]

            "update K, k, V, v"
            # print("q_uu", Q_uu)

            K[t] = -np.dot(inv(Q_uu), Q_ux)
            k[t] = -np.dot(inv(Q_uu), q_u)

            V[t] = Q_xx + np.dot(Q_xu, K[t]) + np.dot(K[t].T, Q_ux) + np.dot(np.dot(K[t].T, Q_uu), K[t])
            v[t] = q_x + np.dot(Q_xu, k[t]) + np.dot(K[t].T, q_u) + np.dot(np.dot(K[t].T, Q_uu), k[t])

            self.std.append(inv(Q_uu))
            t -= 1

        self.K = K
        self.k = k

    def get_action_one_step(self, state, t):

        mean = np.dot(self.K[t], state) + self.k[t]
        # todo remove random here
        if np.isclose(0.0, self.std[t]).all() is False:
            return np.clip(np.random.normal(mean, self.std[t]), self.control_low, self.control_high)
        else:
            return np.clip(mean, self.control_low, self.control_high)


class LQRPolicy(DeterministicPolicy):

    @typechecked
    def __init__(self, env_spec: EnvSpec, T: int, cost_fn: CostFunc,
                 dynamics: LinearDynamicsModel):
        param = Parameters(parameters=dict(T=T))
        super().__init__(env_spec, param)
        self.dynamics = dynamics
        self.Lqr_instance = LQR(env_spec=env_spec,
                                T=self.parameters('T'),
                                dyna_model=dynamics,
                                cost_fn=cost_fn)

    def forward(self, obs, **kwargs):
        obs = make_batch(obs, original_shape=self.env_spec.obs_shape).tolist()
        action = []
        if 'step' in kwargs:
            step = kwargs['step']
        else:
            step = None
        for obs_i in obs:
            action_i = self._forward(obs_i, step=step)
            action.append(action_i)
        return np.array(action)

    def copy_from(self, obj) -> bool:
        super().copy_from(obj)
        self.parameters.copy_from(obj.parameters)
        return True

    def make_copy(self, *args, **kwargs):
        return LQRPolicy(env_spec=self.env_spec,
                         T=self.parameters('T'),
                         cost_fn=self.Lqr_instance.cost_fn,
                         dynamics=self.dynamics.make_copy())

    def init(self, source_obj=None):
        self.parameters.init()
        if source_obj:
            self.copy_from(obj=source_obj)

    def get_status(self):
        return super().get_status()

    def _forward(self, obs, step: None):
        self.Lqr_instance.backward(np.array([obs]), np.array([self.action_space.sample()]))
        x = obs
        u = self.Lqr_instance.get_action_one_step(x, t=step if step else 0)
        return u

    @property
    def T(self):
        return self.parameters('T')



import numpy as np
from baconian.common.spaces.base import Space


class Box(Space):
    """
    A box in R^n.
    I.e., each coordinate is bounded.
    """

    def __init__(self, low, high, shape=None):
        """
        Two kinds of valid input:
            Box(-1.0, 1.0, (3,4)) # low and high are scalars, and shape is
            provided
            Box(np.array([-1.0,-2.0]), np.array([2.0,4.0])) # low and high are
            arrays of the same shape
        """
        if shape is None:
            assert low.shape == high.shape
            self.low = low
            self.high = high
        else:
            assert np.isscalar(low) and np.isscalar(high)
            self.low = low + np.zeros(shape)
            self.high = high + np.zeros(shape)

    def sample(self):
        return np.random.uniform(
            low=self.low, high=self.high,
            size=self.low.shape).astype(np.float32)

    def contains(self, x):
        return bool(x.shape == self.shape and (x >= self.low).all() and (
                x <= self.high).all())

    @property
    def shape(self):
        return self.low.shape

    @property
    def flat_dim(self):
        return np.prod(self.low.shape)

    @property
    def bounds(self):
        return self.low, self.high

    def flatten(self, x):
        return np.asarray(x).flatten()

    def unflatten(self, x):
        return np.asarray(x).reshape(self.shape)

    def flatten_n(self, xs):
        xs = np.asarray(xs)
        return xs.reshape((xs.shape[0], -1))

    def unflatten_n(self, xs):
        xs = np.asarray(xs)
        return xs.reshape((xs.shape[0],) + self.shape)

    def __repr__(self):
        return "Box" + str(self.shape)

    def __eq__(self, other):
        return isinstance(other, Box) \
               and np.allclose(self.low, other.low) \
               and np.allclose(self.high, other.high)

    def __hash__(self):
        return hash((tuple(self.low.tolist()), tuple(self.high.tolist())))

    def new_tensor_variable(self, name, extra_dims):
        raise NotImplementedError

    def clip(self, x):
        return np.clip(x, self.low, self.high).reshape(self.shape)

    def bound(self):
        return self.low, self.high



from baconian.common.sampler.sample_data import TransitionData
from baconian.test.tests.set_up.setup import TestWithAll
import numpy as np
from baconian.algo.dynamics.terminal_func.terminal_func import *
from baconian.algo.dynamics.reward_func.reward_func import RandomRewardFunc


class TestDynamics(TestWithAll):

    def test_init(self):
        ddpg, locals = self.create_ddpg()
        env_spec = locals['env_spec']
        env = locals['env']
        mlp_dyna = self.create_continuous_mlp_global_dynamics_model(env_spec=env_spec)[0]
        algo = self.create_dyna(env_spec=env_spec, model_free_algo=ddpg, dyanmics_model=mlp_dyna)[0]
        algo.init()

        st = env.reset()
        data = TransitionData(env_spec)

        for _ in range(100):
            ac = algo.predict(st)
            new_st, re, done, _ = env.step(action=ac)
            data.append(state=st,
                        new_state=new_st,
                        reward=re,
                        action=ac,
                        done=done)
        algo.append_to_memory(samples=data)
        pre_res = 10000
        for i in range(20):
            print(algo.train(batch_data=data))
            print(algo.train(batch_data=data, state='state_dynamics_training'))
            print(algo.train(batch_data=data, state='state_agent_training'))
            res = algo.test_dynamics(env=env, sample_count=100)
            self.assertLess(list(res.values())[0], pre_res)
            self.assertLess(list(res.values())[1], pre_res)
            print(res)
        algo.test()

    def test_dynamics_as_env(self):
        env = self.create_env('Pendulum-v0')
        env_spec = self.create_env_spec(env)

        mlp_dyna = self.create_continuous_mlp_global_dynamics_model(env_spec=env_spec)[0]
        env = mlp_dyna.return_as_env()
        env.init()
        env.set_terminal_reward_func(terminal_func=FixedEpisodeLengthTerminalFunc(max_step_length=10,
                                                                                  step_count_fn=lambda: env.total_step_count_fn() - env._last_reset_point),
                                     reward_func=RandomRewardFunc())
        env.reset()
        self.assertEqual(env._last_reset_point, 0)
        for i in range(11):
            new_st, re, done, _ = env.step(action=env_spec.action_space.sample())
            self.assertEqual(env.total_step_count_fn(), i + 1)
            if done is True:
                self.assertEqual(i, 9)
                env.reset()
                self.assertEqual(env._last_reset_point, env.total_step_count_fn())
                self.assertEqual(env._last_reset_point, i + 1)



from baconian.algo.dynamics.dynamics_model import DynamicsModel
from typeguard import typechecked
from baconian.config.dict_config import DictConfig
from baconian.common.sampler.sample_data import TransitionData
from baconian.core.parameters import Parameters
from baconian.config.global_config import GlobalConfig
from baconian.algo.rl_algo import ModelFreeAlgo, ModelBasedAlgo
from baconian.common.misc import *
from baconian.common.logging import record_return_decorator
from baconian.core.status import register_counter_info_to_status_decorator
from baconian.core.util import init_func_arg_record_decorator
from baconian.algo.misc.placeholder_input import PlaceholderInput
import os


class Dyna(ModelBasedAlgo):
    """
    Dyna algorithms, Sutton, R. S. (1991).
    You can replace the dynamics model with any dynamics models you want.
    """
    required_key_dict = DictConfig.load_json(file_path=GlobalConfig().DEFAULT_ALGO_DYNA_REQUIRED_KEY_LIST)

    @init_func_arg_record_decorator()
    @typechecked
    def __init__(self, env_spec, dynamics_model: DynamicsModel,
                 model_free_algo: ModelFreeAlgo,
                 config_or_config_dict: (DictConfig, dict),
                 name='sample_with_dynamics'
                 ):
        super().__init__(env_spec, dynamics_model, name)
        config = construct_dict_config(config_or_config_dict, self)
        parameters = Parameters(parameters=dict(),
                                name='dyna_param',
                                source_config=config)
        sub_placeholder_input_list = []
        if isinstance(dynamics_model, PlaceholderInput):
            sub_placeholder_input_list.append(dict(obj=dynamics_model,
                                                   attr_name='dynamics_model'))
        if isinstance(model_free_algo, PlaceholderInput):
            sub_placeholder_input_list.append(dict(obj=model_free_algo,
                                                   attr_name='model_free_algo'))
        self.model_free_algo = model_free_algo
        self.config = config
        self.parameters = parameters

    @register_counter_info_to_status_decorator(increment=1, info_key='init', under_status='INITED')
    def init(self):
        self.parameters.init()
        self.model_free_algo.init()
        self.dynamics_env.init()
        super().init()

    @record_return_decorator(which_recorder='self')
    @register_counter_info_to_status_decorator(increment=1, info_key='train_counter', under_status='TRAIN')
    def train(self, *args, **kwargs) -> dict:
        super(Dyna, self).train()
        res_dict = {}
        batch_data = kwargs['batch_data'] if 'batch_data' in kwargs else None
        if 'state' in kwargs:
            assert kwargs['state'] in ('state_dynamics_training', 'state_agent_training')
            state = kwargs['state']
            kwargs.pop('state')
        else:
            state = None

        if not state or state == 'state_dynamics_training':

            dynamics_train_res_dict = self._fit_dynamics_model(batch_data=batch_data,
                                                               train_iter=self.parameters('dynamics_model_train_iter'))
            for key, val in dynamics_train_res_dict.items():
                res_dict["{}_{}".format(self._dynamics_model.name, key)] = val
        if not state or state == 'state_agent_training':
            model_free_algo_train_res_dict = self._train_model_free_algo(batch_data=batch_data,
                                                                         train_iter=self.parameters(
                                                                             'model_free_algo_train_iter'))

            for key, val in model_free_algo_train_res_dict.items():
                res_dict['{}_{}'.format(self.model_free_algo.name, key)] = val
        return res_dict

    @register_counter_info_to_status_decorator(increment=1, info_key='test_counter', under_status='TEST')
    def test(self, *arg, **kwargs):
        return super().test(*arg, **kwargs)

    @register_counter_info_to_status_decorator(increment=1, info_key='predict_counter')
    def predict(self, obs, **kwargs):
        return self.model_free_algo.predict(obs)

    def append_to_memory(self, *args, **kwargs):
        self.model_free_algo.append_to_memory(kwargs['samples'])

    @record_return_decorator(which_recorder='self')
    def save(self, global_step, save_path=None, name=None, **kwargs):
        save_path = save_path if save_path else GlobalConfig().DEFAULT_MODEL_CHECKPOINT_PATH
        name = name if name else self.name
        self.model_free_algo.save(global_step=global_step,
                                  name=None,
                                  save_path=os.path.join(save_path, self.model_free_algo.name))
        self.dynamics_env.save(global_step=global_step,
                               name=None,
                               save_path=os.path.join(save_path, self.dynamics_env.name))
        return dict(check_point_save_path=save_path, check_point_save_global_step=global_step,
                    check_point_save_name=name)

    @record_return_decorator(which_recorder='self')
    def load(self, path_to_model, model_name, global_step=None, **kwargs):
        self.model_free_algo.load(path_to_model=os.path.join(path_to_model, self.model_free_algo.name),
                                  model_name=self.model_free_algo.name,
                                  global_step=global_step)
        self.dynamics_env.load(global_step=global_step,
                               path_to_model=os.path.join(path_to_model, self.dynamics_env.name),
                               model_name=self.dynamics_env.name)
        return dict(check_point_load_path=path_to_model, check_point_load_global_step=global_step,
                    check_point_load_name=model_name)

    @register_counter_info_to_status_decorator(increment=1, info_key='dyanmics_train_counter', under_status='TRAIN')
    def _fit_dynamics_model(self, batch_data: TransitionData, train_iter, sess=None) -> dict:
        res_dict = self._dynamics_model.train(batch_data, **dict(sess=sess,
                                                                 train_iter=train_iter))
        return res_dict

    @register_counter_info_to_status_decorator(increment=1, info_key='mode_free_algo_dyanmics_train_counter',
                                               under_status='TRAIN')
    def _train_model_free_algo(self, batch_data=None, train_iter=None, sess=None):
        res_dict = self.model_free_algo.train(**dict(batch_data=batch_data,
                                                     train_iter=train_iter,
                                                     sess=sess))
        return res_dict



from baconian.core.core import Basic, Env
from baconian.algo.dynamics.dynamics_model import DynamicsModel, DynamicsEnvWrapper
from baconian.algo.algo import Algo
import numpy as np
import abc


class Ensemble(Basic):

    def init(self, *args, **kwargs):
        raise NotImplementedError

    def save(self, *args, **kwargs):
        raise NotImplementedError

    def load(self, *args, **kwargs):
        raise NotImplementedError


class ModelEnsemble(Ensemble, DynamicsModel):

    def __init__(self, model, n_models=1, prediction_type='random', *args, **kwargs):
        """
        :param model:
        :param n_models:
        :param prediction_type:
        """

        super().__init__(*args, **kwargs)

        self._prediction_type = prediction_type
        self._observations = list()
        self._model = list()
        self._name = kwargs.pop('name', 'dynamics_model')

        if isinstance(model, DynamicsModel):
            for a in range(n_models):
                self._model.append(model.make_copy())
        else:
            base_env_spec = model[0].env_spec
            for b in range(len(model)):
                if model[b].env_spec != base_env_spec:
                    raise TypeError('EnvSpec of list of models do not match.')
            self._model = model
        self.state = None

    def train(self, *args, **kwargs):
        res = {}
        for idx in range(len(self._model)):
            res['model_{}'.format(idx)] = self._model[idx].train(*args, **kwargs)
        return res

    def reset_state(self, *args, **kwargs):
        """
        Reset the model parameters.
        """
        self._observations = list()

        for m in self.model:
            m.reset_state(*args, **kwargs)
            self._observations.append(m.state)
        if self._prediction_type == 'mean':
            self.state = np.mean(self._observations, axis=0)
        elif self._prediction_type == 'random':
            self.state = self._observations[np.random.randint(low=0, high=len(self.model))]
        else:
            raise ValueError

    def init(self, *args, **kwargs):

        for m in self.model:
            m.init(*args, **kwargs)
        self.reset_state()

    def step(self, *args, **kwargs):

        self._observations = list()

        for m in self.model:
            self._observations.append(m.step(*args, **kwargs))

        if self._prediction_type == 'mean':
            results = np.mean(self._observations, axis=0)
        elif self._prediction_type == 'random':
            results = self._observations[np.random.randint(low=0, high=len(self.model))]
        else:
            raise ValueError

        return results

    def get_obs(self):
        return self._observations

    @abc.abstractmethod
    def _state_transit(self, state, action, **kwargs) -> np.ndarray:
        raise NotImplementedError

    def copy_from(self, obj) -> bool:
        if not isinstance(obj, type(self)):
            raise TypeError('Wrong type of obj %s to be copied, which should be %s' % (type(obj), type(self)))
        return True

    def make_copy(self):
        raise NotImplementedError

    def return_as_env(self) -> Env:
        return DynamicsEnvWrapper(dynamics=self,
                                  name=self._name + '_env')

    def save(self, *args, **kwargs):
        for model in self._model:
            model.save(*args, **kwargs)

    def load(self, *args, **kwargs):
        raise NotImplementedError

    @property
    def model(self):
        """
        Returns:
            The list of the models in the ensemble.
        """
        return self._model

    def __len__(self):
        return len(self._model)

    def __getitem__(self, idx):
        return self._model[idx]


class AlgoEnsemble(Ensemble, Algo):

    def __init__(self, cls, n_algos, prediction_type='random', *args, **kwargs):

        super().__init__(*args, **kwargs)
        self._cls = cls
        self._n_algos = n_algos
        self._prediction_type = prediction_type
        self._predictions = list()
        self._algo = list()
        self._name = kwargs.pop('name', 'algo')

        for num in range(n_algos):
            self._algo.append(cls(*args, **dict(kwargs, name=self._name + '_' + num)))

    def init(self, *args, **kwargs):
        raise NotImplementedError

    def save(self, *args, **kwargs):
        raise NotImplementedError

    def load(self, *args, **kwargs):
        raise NotImplementedError

    @property
    def algo(self):
        """
        Returns:
            The list of the models in the ensemble.
        """
        return self._algo

    def __len__(self):
        return len(self._algo)

    def __getitem__(self, idx):
        return self._algo[idx]



from baconian.common.logging import Recorder


class Tuner(object):
    """
    Auto hyper parameter tuning module, tobe done
    """
    def __init__(self):
        self.recorder = Recorder(default_obj=self)



from copy import deepcopy
from collections import Hashable
from baconian.common.error import *
from baconian.core.global_var import get_all, reset
from functools import wraps
from baconian.config.global_config import GlobalConfig


def init_func_arg_record_decorator():
    def wrap(fn):
        @wraps(fn)
        def wrap_with_self(self, *args, **kwargs):
            get_all()['_global_obj_arg_dict'][self] = None
            # get_all()['_global_obj_arg_dict'][self] = dict(args=args, kwargs=kwargs, cls=type(self))
            res = fn(self, *args, **kwargs)
            return res

        return wrap_with_self

    return wrap


def get_global_arg_dict():
    return get_all()['_global_obj_arg_dict']


def copy_globally(arg_dict, source_obj_list):
    """
    deprecated in the future
    :param arg_dict:
    :param source_obj_list:
    :return:
    """
    new_obj_list = []
    reset('_global_name_dict')
    for obj in source_obj_list:
        if obj not in arg_dict:
            raise ValueError('{} not in arg_dict'.format(obj))
        else:
            new_obj_list.append(_make_copy_object(arg_dict, obj=obj))

    return new_obj_list


def register_name_globally(name: str, obj):
    if name in get_all()['_global_name_dict'] and not id(obj) == id(
            get_all()['_global_name_dict'][name]) and obj.allow_duplicate_name is False and \
            get_all()['_global_name_dict'][
                name].allow_duplicate_name is False and GlobalConfig().DEFAULT_TURN_OFF_GLOBAL_NAME_FLAG is False:
        raise GlobalNameExistedError(
            'name : {} is existed with object: {}'.format(name, get_all()['_global_name_dict'][name]))
    else:
        get_all()['_global_name_dict'][name] = obj


def _make_copy_object(arg_dict: dict, obj):
    if obj not in arg_dict:
        raise ValueError('{} not in arg_dict'.format(obj))
    else:
        args = arg_dict[obj]['args']
        kwargs = arg_dict[obj]['kwargs']
        cls = arg_dict[obj]['cls']

        new_args = []
        new_kwargs = dict()

        arg_dict.pop(obj)
        del obj
        for a in args:
            if in_dict(a, args):
                new_args.append(_make_copy_object(arg_dict, obj=a))
            else:
                new_args.append(deepcopy(a))
        for key, a in kwargs.items():
            print(key, a)
            if in_dict(a, arg_dict):
                new_kwargs[key] = _make_copy_object(arg_dict, obj=a)
            else:
                new_kwargs[key] = deepcopy(a)
        print("create ", cls, new_args, new_kwargs, flush=True)
        return cls(*new_args, **new_kwargs)


def in_dict(obj, list_or_dict):
    if isinstance(obj, Hashable):
        return obj in list_or_dict
    else:
        if isinstance(list_or_dict, (tuple, list)):
            for key in list_or_dict:
                if id(key) == id(obj):
                    return True
            return False
        elif isinstance(list_or_dict, dict):
            for key, val in list_or_dict.items():
                if id(key) == id(obj):
                    return True
            return False



# import numpy as np
from scipy.linalg import inv
from scipy.optimize import approx_fprime

from baconian.algo.policy.policy import DeterministicPolicy
from baconian.core.parameters import Parameters
from baconian.core.core import EnvSpec
import autograd.numpy as np
from baconian.common.special import *
from baconian.algo.dynamics.reward_func.reward_func import CostFunc
from baconian.algo.rl_algo import ModelBasedAlgo
from baconian.common.logging import record_return_decorator
from baconian.core.status import register_counter_info_to_status_decorator
from baconian.common.sampler.sample_data import TransitionData

from baconian.algo.dynamics.dynamics_model import DynamicsEnvWrapper

"""
the gradient is computed approximated instead of analytically
"""


class iLQR(object):

    def __init__(self, env_spec: EnvSpec, delta, T, dyn_model: DynamicsEnvWrapper, cost_fn: CostFunc):

        self.env_spec = env_spec
        self.min_factor = 2
        self.factor = self.min_factor
        self.min_mu = 1e-6
        self.mu = self.min_mu
        self.delta = delta
        self.T = T
        self.dyn_model = dyn_model
        self.cost_fn = cost_fn
        self.control_low = self.env_spec.action_space.low
        self.control_high = self.env_spec.action_space.high
        self.K, self.k, self.std = None, None, None

    def increase(self, mu):
        self.factor = np.maximum(self.factor, self.factor * self.min_factor)
        self.mu = np.maximum(self.min_mu, self.mu * self.factor)

    def decrease(self, mu):
        self.factor = np.minimum(1 / self.min_factor, self.factor / self.min_factor)
        if self.mu * self.factor > self.min_mu:
            self.mu = self.mu * self.factor
        else:
            self.mu = 0

    def simulate_step(self, x):

        xu = [x[:self.env_spec.obs_space.shape[0]], x[self.env_spec.obs_space.shape[0]:]]

        next_x = self.dyn_model.step(state=xu[0], action=xu[1], allow_clip=True)
        # next_x = self.env_spec.obs_space.sample()
        "get cost"
        cost = self.cost_fn(state=xu[0], action=xu[1], new_state=next_x)

        return next_x, cost

    def simulate_next_state(self, x, i):
        return self.simulate_step(x)[0][i]

    def simulate_cost(self, x):
        return self.simulate_step(x)[1]

    def approx_fdoubleprime(self, x, i):
        return approx_fprime(x, self.simulate_cost, self.delta)[i]

    def finite_difference(self, x, u):

        "calling finite difference for delta perturbation"
        xu = np.concatenate((x, u))

        F = np.zeros((x.shape[0], xu.shape[0]))

        for i in range(x.shape[0]):
            F[i, :] = approx_fprime(xu, self.simulate_next_state, self.delta, i)

        c = approx_fprime(xu, self.simulate_cost, self.delta)

        C = np.zeros((len(xu), len(xu)))

        for i in range(xu.shape[0]):
            C[i, :] = approx_fprime(xu, self.approx_fdoubleprime, self.delta, i)

        f = np.zeros((len(x)))

        return C, F, c, f

    def differentiate(self, x_seq, u_seq):

        "get gradient values using finite difference"

        C, F, c, f = [], [], [], []

        for t in range(self.T - 1):
            Ct, Ft, ct, ft = self.finite_difference(x_seq[t], u_seq[t])

            C.append(Ct)
            F.append(Ft)
            c.append(ct)
            f.append(ft)

        "TODO : C, F, c, f for time step T are different. Why ?"

        u = np.zeros((u_seq[0].shape))

        Ct, Ft, ct, ft = self.finite_difference(x_seq[-1], u)

        C.append(Ct)
        F.append(Ft)
        c.append(ct)
        f.append(ft)

        return C, F, c, f

    def backward(self, x_seq, u_seq):

        "initialize F_t, C_t, f_t, c_t, V_t, v_t"
        C, F, c, f = self.differentiate(x_seq, u_seq)

        n = x_seq[0].shape[0]

        "initialize V_t1 and v_t1"

        c_x = c[-1][:n]
        c_u = c[-1][n:]

        C_xx = C[-1][:n, :n]
        C_xu = C[-1][:n, n:]
        C_ux = C[-1][n:, :n]
        C_uu = C[-1][n:, n:]

        # C_uu = C_uu + self.mu * np.eye(C_uu.shape[0])

        K = np.zeros((self.T + 1, u_seq[0].shape[0], x_seq[0].shape[0]))
        k = np.zeros((self.T + 1, u_seq[0].shape[0]))

        V = np.zeros((self.T + 1, x_seq[0].shape[0], x_seq[0].shape[0]))
        v = np.zeros((self.T + 1, x_seq[0].shape[0]))

        # K[-1] = -np.dot(inv(C_uu), C_ux)
        # k[-1] = -np.dot(inv(C_uu), c_u)

        # V[-1] = C_xx + np.dot(C_xu, K[-1]) + np.dot(K[-1].T, C_ux) + np.dot(np.dot(K[-1].T, C_uu), K[-1])
        # v[-1] = c_x + np.dot(C_xu, k[-1]) + np.dot(K[-1].T, c_u) + np.dot(np.dot(K[-1].T, C_uu), k[-1])
        V[-1] = C_xx
        v[-1] = c_x

        "initialize Q_t1 and q_t1"

        Q = list(np.zeros((self.T)))
        q = list(np.zeros((self.T)))

        self.std = []
        "loop till horizon"

        t = self.T - 1
        while t >= 0:
            # for t in range(self.T-1, -1, -1):
            "update Q"
            # todo remove mu here
            # Q[t] = C[t] + np.dot(np.dot(F[t].T, V[t + 1] + self.mu * np.eye(V[t + 1].shape[0])),
            #                      F[t])  # + 0.01 * np.eye(V[t+1].shape[0])),

            Q[t] = C[t] + np.dot(np.dot(F[t].T, V[t + 1]), F[t])
            q[t] = c[t] + np.dot(F[t].T, v[t + 1]) + np.dot(np.dot(F[t].T, V[t + 1]), f[t])

            "differentiate Q to get Q_uu, Q_xx, Q_ux, Q_u, Q_x"

            q_x = q[t][:n]
            q_u = q[t][n:]

            Q_xx = Q[t][:n, :n]
            Q_xu = Q[t][:n, n:]
            Q_ux = Q[t][n:, :n]
            Q_uu = Q[t][n:, n:]

            # Q_uu = Q_uu + 100 * np.eye(Q_uu.shape[0])
            # try:
            #     np.linalg.cholesky(Q_uu)
            # except:
            #     print(self.mu)
            #     self.increase(self.mu)
            #     continue

            "update K, k, V, v"
            # print("q_uu", Q_uu)

            K[t] = -np.dot(inv(Q_uu), Q_ux)
            k[t] = -np.dot(inv(Q_uu), q_u)

            V[t] = Q_xx + np.dot(Q_xu, K[t]) + np.dot(K[t].T, Q_ux) + np.dot(np.dot(K[t].T, Q_uu), K[t])
            v[t] = q_x + np.dot(Q_xu, k[t]) + np.dot(K[t].T, q_u) + np.dot(np.dot(K[t].T, Q_uu), k[t])

            self.std.append(inv(Q_uu))
            # self.decrease(self.mu)
            t -= 1

        self.K = K
        self.k = k

    def get_action_one_step(self, state, t, x, u):

        "TODO : Add delta U's to given action array"
        # print("Q_uu ", self.std[t])
        mean = np.dot(self.K[t], (state - x)) + self.k[t] + u
        # return np.clip(mean, self.control_low, self.control_high)
        # todo remove random here
        if np.isclose(0.0, self.std[t]).all() is False:
            return np.clip(np.random.normal(mean, self.std[t]), self.control_low, self.control_high)
        else:
            return np.clip(mean, self.control_low, self.control_high)


class iLQRPolicy(DeterministicPolicy):

    @typechecked
    def __init__(self, env_spec: EnvSpec, T: int, delta: float, iteration: int, cost_fn: CostFunc,
                 dynamics_model_train_iter: int,
                 dynamics: DynamicsEnvWrapper):
        param = Parameters(parameters=dict(T=T, delta=delta,
                                           iteration=iteration,
                                           dynamics_model_train_iter=dynamics_model_train_iter))
        super().__init__(env_spec, param)
        self.dynamics = dynamics
        self.U_hat = None
        self.X_hat = None
        self.iLqr_instance = iLQR(env_spec=env_spec,
                                  delta=self.parameters('delta'),
                                  T=self.parameters('T'),
                                  dyn_model=dynamics._dynamics,
                                  cost_fn=cost_fn)

    def forward(self, obs, **kwargs):
        obs = make_batch(obs, original_shape=self.env_spec.obs_shape).tolist()
        action = []
        if 'step' in kwargs:
            step = kwargs['step']
        else:
            step = None
        for obs_i in obs:
            action_i = self._forward(obs_i, step=step)
            action.append(action_i)
        return np.array(action)

    def copy_from(self, obj) -> bool:
        super().copy_from(obj)
        self.parameters.copy_from(obj.parameters)
        return True

    def make_copy(self, *args, **kwargs):
        dynamics = DynamicsEnvWrapper(dynamics=self.dynamics._dynamics)
        dynamics.set_terminal_reward_func(terminal_func=self.dynamics._terminal_func,
                                          reward_func=self.dynamics._reward_func)
        return iLQRPolicy(env_spec=self.env_spec,
                          T=self.parameters('T'),
                          delta=self.parameters('delta'),
                          iteration=self.parameters('iteration'),
                          cost_fn=self.iLqr_instance.cost_fn,
                          dynamics_model_train_iter=self.parameters('dynamics_model_train_iter'),
                          dynamics=dynamics)

    def init(self, source_obj=None):
        self.parameters.init()
        if source_obj:
            self.copy_from(obj=source_obj)

    def get_status(self):
        return super().get_status()

    def _forward(self, obs, step: None):
        if not step:
            self.U_hat = np.reshape([np.zeros(self.action_space.sample().shape) for _ in range(self.T - 1)],
                                    (self.T - 1, self.action_space.shape[0]))

            self.X_hat = []
            self.X_hat.append(obs)
            x = obs
            for i in range(self.T - 1):
                next_obs, _, _, _ = self.dynamics.step(action=self.U_hat[i, :], state=x, allow_clip=True)
                self.X_hat.append(next_obs)
                x = next_obs
            self.X_hat = np.array(self.X_hat)
        for i in range(self.parameters('iteration')):
            self.iLqr_instance.backward(self.X_hat, self.U_hat)
            x = obs
            U = np.zeros(self.U_hat.shape)
            X = np.zeros(self.X_hat.shape)

            for t in range(self.T - 1):
                u = self.iLqr_instance.get_action_one_step(x, t, self.X_hat[t], self.U_hat[t])
                X[t] = x
                U[t] = u
                x, _, _, _ = self.dynamics.step(state=x, action=u, allow_clip=True)

            X[-1] = x
            self.X_hat = X
            self.U_hat = U
        return self.U_hat[0]

    @property
    def T(self):
        return self.parameters('T')


class iLQRAlogWrapper(ModelBasedAlgo):

    def __init__(self, policy, env_spec, dynamics_env: DynamicsEnvWrapper, name: str = 'model_based_algo'):
        self.policy = policy
        super().__init__(env_spec, dynamics_env._dynamics, name)
        self.dynamics_env = dynamics_env

    def predict(self, obs, **kwargs):
        if self.is_training is True:
            return self.env_spec.action_space.sample()
        else:
            return self.policy.forward(obs=obs)

    def append_to_memory(self, *args, **kwargs):
        pass

    def init(self):
        self.policy.init()
        self.dynamics_env.init()
        super().init()

    @record_return_decorator(which_recorder='self')
    @register_counter_info_to_status_decorator(increment=1, info_key='train_counter', under_status='TRAIN')
    def train(self, *args, **kwargs) -> dict:
        super(iLQRAlogWrapper, self).train()
        res_dict = {}
        batch_data = kwargs['batch_data'] if 'batch_data' in kwargs else None
        if 'state' in kwargs:
            assert kwargs['state'] in ('state_dynamics_training', 'state_agent_training')
            state = kwargs['state']
            kwargs.pop('state')
        else:
            state = None

        if not state or state == 'state_dynamics_training':

            dynamics_train_res_dict = self._fit_dynamics_model(batch_data=batch_data,
                                                               train_iter=self.policy.parameters(
                                                                   'dynamics_model_train_iter'))
            for key, val in dynamics_train_res_dict.items():
                res_dict["{}_{}".format(self._dynamics_model.name, key)] = val
        return res_dict

    @register_counter_info_to_status_decorator(increment=1, info_key='dyanmics_train_counter', under_status='TRAIN')
    def _fit_dynamics_model(self, batch_data: TransitionData, train_iter, sess=None) -> dict:
        res_dict = self._dynamics_model.train(batch_data, **dict(sess=sess,
                                                                 train_iter=train_iter))
        return res_dict



from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
import numpy as np

env = make("HalfCheetah-v2")
env_spec = EnvSpec(obs_space=env.observation_space, action_space=env.action_space)

OBS_DIM = env_spec.flat_obs_dim
HID1_SIZE = 400
HID2_SIZE = 300

POLICY_HID_MULTI = 10
ACT_DIM = env_spec.flat_action_dim
POLICY_HID1_SIZE = 400
POLICY_HID2_SIZE = 300

CHEETAH_BENCHMARK_CONFIG_DICT = {
    'env_id': "HalfCheetah-v2",
    'MLP_V': {
        'mlp_config': [
            {
                "ACT": "RELU",
                "B_INIT_VALUE": 0.0,
                "NAME": "1",
                "N_UNITS": HID1_SIZE,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": np.sqrt(1 / OBS_DIM)
            },
            {
                "ACT": "RELU",
                "B_INIT_VALUE": 0.0,
                "NAME": "2",
                "N_UNITS": HID2_SIZE,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": np.sqrt(1 / HID1_SIZE)
            },
            {
                "ACT": "IDENTITY",
                "B_INIT_VALUE": 0.0,
                "NAME": "OUPTUT",
                "N_UNITS": 1,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": np.sqrt(1 / HID2_SIZE),
            }
        ]
    },
    'POLICY': {
        'mlp_config': [
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "1",
                "N_UNITS": POLICY_HID1_SIZE,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": np.sqrt(1 / OBS_DIM)
            },
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "2",
                "N_UNITS": POLICY_HID2_SIZE,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": np.sqrt(1 / POLICY_HID1_SIZE)
            },
            {
                "ACT": "IDENTITY",
                "B_INIT_VALUE": 0.0,
                "NAME": "OUPTUT",
                "N_UNITS": ACT_DIM,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": np.sqrt(1 / POLICY_HID2_SIZE)
            }
        ]
    },
    'DDPG': {
        'config_or_config_dict': {
            "REPLAY_BUFFER_SIZE": 1000000,
            "GAMMA": 0.99,
            "CRITIC_LEARNING_RATE": 0.001,
            "ACTOR_LEARNING_RATE": 0.0001,
            "DECAY": 0.999,
            "BATCH_SIZE": 128,
            "TRAIN_ITERATION": 120,
            "critic_clip_norm": None,
            "actor_clip_norm": None,
        },
        'replay_buffer': None
    },
    'TrainTestFlow': {
        "TEST_SAMPLES_COUNT": 10,
        "TRAIN_SAMPLES_COUNT": 1,
        'config_or_config_dict': {
            "TEST_EVERY_SAMPLE_COUNT": 1000,
            "TRAIN_EVERY_SAMPLE_COUNT": 1,
            "START_TRAIN_AFTER_SAMPLE_COUNT": 10000,
            "START_TEST_AFTER_SAMPLE_COUNT": 20,
        }
    },

    'DEFAULT_EXPERIMENT_END_POINT': dict(TOTAL_AGENT_TRAIN_SAMPLE_COUNT=1000000,
                                         TOTAL_AGENT_TEST_SAMPLE_COUNT=None,
                                         TOTAL_AGENT_UPDATE_COUNT=None),
    'AGENT_NOISE': {

    }
}


from baconian.core.core import Basic, EnvSpec
import typeguard as tg
from baconian.core.parameters import Parameters
import abc


class Policy(Basic):

    @tg.typechecked
    def __init__(self, env_spec: EnvSpec, parameters: Parameters = None, name='policy'):
        super().__init__(name=name)
        self.env_spec = env_spec
        self.parameters = parameters

    @property
    def obs_space(self):
        return self.env_spec.obs_space

    @property
    def action_space(self):
        return self.env_spec.action_space

    @abc.abstractmethod
    def forward(self, *args, **kwargs):
        raise NotImplementedError

    @tg.typechecked
    def copy_from(self, obj) -> bool:
        if not isinstance(obj, type(self)):
            raise TypeError('Wrong type of obj %s to be copied, which should be %s' % (type(obj), type(self)))
        return True

    def make_copy(self, *args, **kwargs):
        raise NotImplementedError

    def init(self, source_obj=None):
        if self.parameters:
            self.parameters.init()
        if source_obj:
            self.copy_from(obj=source_obj)


class StochasticPolicy(Policy):
    @tg.typechecked
    def __init__(self, env_spec: EnvSpec, parameters: Parameters = None, name: str = 'stochastic_policy'):
        super(StochasticPolicy, self).__init__(env_spec=env_spec, parameters=parameters, name=name)
        self.state_input = None
        self.action_output = None

    def log_prob(self, *args, **kwargs):
        pass

    def prob(self, *args, **kwargs):
        pass

    def kl(self, other, *kwargs):
        pass

    def entropy(self, *args, **kwargs):
        pass

    def get_dist_info(self) -> tuple:
        pass


class DeterministicPolicy(Policy):
    pass



import unittest
import tensorflow as tf
from baconian.algo.dynamics.mlp_dynamics_model import ContinuousMLPGlobalDynamicsModel
from baconian.algo.misc.placeholder_input import PlaceholderInput
from baconian.algo.dqn import DQN
from baconian.tf.tf_parameters import ParametersWithTensorflowVariable
from baconian.core.core import Basic, EnvSpec
from baconian.envs.gym_env import make
from baconian.algo.value_func.mlp_q_value import MLPQValueFunction
from baconian.algo.ddpg import DDPG
from baconian.algo.policy import DeterministicMLPPolicy
from baconian.algo.policy import ConstantActionPolicy
from baconian.algo.value_func import MLPVValueFunc
from baconian.algo.policy.normal_distribution_mlp import NormalDistributionMLPPolicy
from baconian.algo.ppo import PPO
from baconian.core.parameters import Parameters, DictConfig
from baconian.algo.mpc import ModelPredictiveControl
from baconian.algo.dynamics.terminal_func.terminal_func import RandomTerminalFunc
from baconian.algo.dynamics.reward_func.reward_func import RandomRewardFunc, CostFunc
from baconian.algo.policy import UniformRandomPolicy
from baconian.core.agent import Agent
from baconian.algo.misc import EpsilonGreedy
from baconian.core.experiment import Experiment
from baconian.core.flow.train_test_flow import TrainTestFlow
from baconian.algo.dyna import Dyna
from baconian.common.schedules import *
from baconian.core.status import *
from baconian.algo.policy.ilqr_policy import iLQRPolicy
from baconian.algo.dynamics.random_dynamics_model import UniformRandomDynamicsModel
from baconian.common.noise import *
from baconian.core.flow.dyna_flow import DynaFlow
from baconian.common.data_pre_processing import *
from baconian.common.sampler.sample_data import TransitionData, TrajectoryData


class Foo(Basic):
    def __init__(self):
        super().__init__(name='foo')

    required_key_dict = dict(var1=1, var2=0.1)


class ClassCreatorSetup(unittest.TestCase):

    def create_env(self, env_id):
        return make(env_id)

    def create_env_spec(self, env):
        return EnvSpec(action_space=env.action_space,
                       obs_space=env.observation_space)

    def create_tf_parameters(self, name='test_tf_param'):
        with tf.variable_scope(name):
            a = tf.get_variable(shape=[3, 4], dtype=tf.float32, name='var_1')
            b = tf.get_variable(shape=[3, 4], dtype=tf.bool, name='var_2')

        conf = DictConfig(required_key_dict=Foo.required_key_dict,
                          config_dict=dict(var1=1, var2=0.01))
        param = ParametersWithTensorflowVariable(tf_var_list=[a, b],
                                                 rest_parameters=dict(var3='sss'),
                                                 name=name,
                                                 source_config=conf,
                                                 require_snapshot=True,
                                                 to_ph_parameter_dict=dict(
                                                     var1=tf.placeholder(shape=(), dtype=tf.int32)))
        return param, locals()

    def create_mlp_q_func(self, env_id='Acrobot-v1', name='mlp_q'):
        env = make(env_id)
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)

        mlp_q = MLPQValueFunction(env_spec=env_spec,
                                  name_scope=name,
                                  name=name,
                                  mlp_config=[
                                      {
                                          "ACT": "RELU",
                                          "B_INIT_VALUE": 0.0,
                                          "NAME": "1",
                                          "N_UNITS": 16,
                                          "TYPE": "DENSE",
                                          "W_NORMAL_STDDEV": 0.03,
                                          "L1_NORM": 0.2,
                                          "L2_NORM": 0.1
                                      },
                                      {
                                          "ACT": "LINEAR",
                                          "B_INIT_VALUE": 0.0,
                                          "NAME": "OUPTUT",
                                          "N_UNITS": 1,
                                          "TYPE": "DENSE",
                                          "W_NORMAL_STDDEV": 0.03
                                      }
                                  ])
        return mlp_q, locals()

    def create_dqn(self, env_id='Acrobot-v1', name='dqn'):
        mlp_q, local = self.create_mlp_q_func(env_id, name='{}_mlp_q'.format(name))
        env_spec = local['env_spec']
        env = local['env']
        dqn = DQN(env_spec=env_spec,
                  config_or_config_dict=dict(REPLAY_BUFFER_SIZE=1000,
                                             GAMMA=0.99,
                                             BATCH_SIZE=10,
                                             LEARNING_RATE=0.001,
                                             TRAIN_ITERATION=1,
                                             DECAY=0.5),
                  name=name,
                  value_func=mlp_q)
        return dqn, locals()

    def create_ph(self, name):
        with tf.variable_scope(name):
            a = tf.get_variable(shape=[3, 4], dtype=tf.float32, name='var_1')

        conf = DictConfig(required_key_dict=Foo.required_key_dict,
                          config_dict=dict(var1=1, var2=0.01))
        param = ParametersWithTensorflowVariable(tf_var_list=[a],
                                                 rest_parameters=dict(var3='sss'),
                                                 name=name,
                                                 source_config=conf,
                                                 require_snapshot=True,
                                                 to_ph_parameter_dict=dict(
                                                     var1=tf.placeholder(shape=(), dtype=tf.int32)))
        param.init()
        a = PlaceholderInput(parameters=param)

        return a, locals()

    def create_ddpg(self, env_id='Pendulum-v0', name='ddpg'):
        env = make(env_id)
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)

        mlp_q = MLPQValueFunction(env_spec=env_spec,
                                  name_scope=name + 'mlp_q',
                                  name=name + 'mlp_q',
                                  mlp_config=[
                                      {
                                          "ACT": "RELU",
                                          "B_INIT_VALUE": 0.0,
                                          "NAME": "1",
                                          "N_UNITS": 16,
                                          "TYPE": "DENSE",
                                          "W_NORMAL_STDDEV": 0.03
                                      },
                                      {
                                          "ACT": "LINEAR",
                                          "B_INIT_VALUE": 0.0,
                                          "NAME": "OUPTUT",
                                          "N_UNITS": 1,
                                          "TYPE": "DENSE",
                                          "W_NORMAL_STDDEV": 0.03
                                      }
                                  ])
        self.assertTrue(len(mlp_q.parameters('tf_var_list')) == 4)
        policy = DeterministicMLPPolicy(env_spec=env_spec,
                                        name_scope=name + 'mlp_policy',
                                        name=name + 'mlp_policy',
                                        mlp_config=[
                                            {
                                                "ACT": "RELU",
                                                "B_INIT_VALUE": 0.0,
                                                "NAME": "1",
                                                "N_UNITS": 16,
                                                "TYPE": "DENSE",
                                                "W_NORMAL_STDDEV": 0.03
                                            },
                                            {
                                                "ACT": "LINEAR",
                                                "B_INIT_VALUE": 0.0,
                                                "NAME": "OUPTUT",
                                                "N_UNITS": env_spec.flat_action_dim,
                                                "TYPE": "DENSE",
                                                "W_NORMAL_STDDEV": 0.03
                                            }
                                        ],
                                        reuse=False)
        self.assertTrue(len(policy.parameters('tf_var_list')) == 4)

        ddpg = DDPG(
            env_spec=env_spec,
            config_or_config_dict={
                "REPLAY_BUFFER_SIZE": 10000,
                "GAMMA": 0.999,
                "CRITIC_LEARNING_RATE": 0.001,
                "ACTOR_LEARNING_RATE": 0.001,
                "DECAY": 0.5,
                "BATCH_SIZE": 50,
                "TRAIN_ITERATION": 1,
                "critic_clip_norm": 0.1,
                "actor_clip_norm": 0.1,
            },
            value_func=mlp_q,
            policy=policy,
            name=name,
            replay_buffer=None
        )
        return ddpg, locals()

    def create_mlp_v(self, env_id='Pendulum-v0', name='mlp_v'):
        env = make(env_id)
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)

        mlp_v = MLPVValueFunc(env_spec=env_spec,
                              name_scope=name + 'mlp_v',
                              name=name + 'mlp_v',
                              mlp_config=[
                                  {
                                      "ACT": "RELU",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "1",
                                      "N_UNITS": 16,
                                      "L1_NORM": 0.01,
                                      "L2_NORM": 0.01,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  },
                                  {
                                      "ACT": "LINEAR",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "OUPTUT",
                                      "N_UNITS": 1,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  }
                              ])
        return mlp_v, locals()

    def create_normal_dist_mlp_policy(self, env_spec, name='norm_dist_p_'):
        policy = NormalDistributionMLPPolicy(env_spec=env_spec,
                                             name_scope=name + 'mlp_policy',
                                             name=name + 'mlp_policy',
                                             mlp_config=[
                                                 {
                                                     "ACT": "RELU",
                                                     "B_INIT_VALUE": 0.0,
                                                     "NAME": "1",
                                                     "L1_NORM": 0.01,
                                                     "L2_NORM": 0.01,
                                                     "N_UNITS": 16,
                                                     "TYPE": "DENSE",
                                                     "W_NORMAL_STDDEV": 0.03
                                                 },
                                                 {
                                                     "ACT": "LINEAR",
                                                     "B_INIT_VALUE": 0.0,
                                                     "NAME": "OUPTUT",
                                                     "N_UNITS": env_spec.flat_action_dim,
                                                     "TYPE": "DENSE",
                                                     "W_NORMAL_STDDEV": 0.03
                                                 }
                                             ],
                                             reuse=False)
        return policy, locals()

    def create_ppo(self, env_id='Pendulum-v0', name='ppo'):
        mlp_v, local = self.create_mlp_v(env_id, name)
        env_spec = local['env_spec']
        env = local['env']
        policy = self.create_normal_dist_mlp_policy(env_spec=env_spec, name=name)[0]
        ppo = PPO(
            env_spec=env_spec,
            config_or_config_dict={
                "gamma": 0.995,
                "lam": 0.98,
                "policy_train_iter": 10,
                "value_func_train_iter": 10,
                "clipping_range": None,
                "beta": 1.0,
                "eta": 50,
                "log_var_init": -1.0,
                "kl_target": 0.003,
                "policy_lr": 0.01,
                "value_func_lr": 0.01,
                "value_func_train_batch_size": 10,
                "lr_multiplier": 1.0
            },
            value_func=mlp_v,
            stochastic_policy=policy,
            name=name
        )
        return ppo, locals()

    def create_dict_config(self):
        a = DictConfig(required_key_dict=Foo.required_key_dict,
                       config_dict=dict(var1=1, var2=0.1),
                       cls_name='Foo')
        return a, locals()

    def create_parameters(self):
        parameters = dict(param1='aaaa',
                          param2=12312,
                          param3=np.random.random([4, 2]))
        source_config, _ = self.create_dict_config()
        a = Parameters(parameters=parameters, source_config=source_config,
                       name='test_params')
        return a, locals()

    def create_continue_dynamics_model(self, env_id='Acrobot-v1', name='mlp_dyna'):
        env = make(env_id)
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)

        mlp_dyna, _ = self.create_continuous_mlp_global_dynamics_model(env_spec=env_spec, name=name)
        return mlp_dyna, locals()

    def create_mpc(self, env_id='Acrobot-v1', name='mpc', policy=None, mlp_dyna=None, env_spec=None, env=None):
        if mlp_dyna is None:
            mlp_dyna, local = self.create_continue_dynamics_model(env_id, name + 'mlp_dyna')
            env_spec = local['env_spec']
            env = local['env']

        policy = policy if policy else UniformRandomPolicy(env_spec=env_spec, name='unp')

        algo = ModelPredictiveControl(
            dynamics_model=mlp_dyna,
            env_spec=env_spec,
            config_or_config_dict=dict(
                SAMPLED_HORIZON=2,
                SAMPLED_PATH_NUM=5,
                dynamics_model_train_iter=10
            ),
            name=name,
            policy=policy
        )
        algo.set_terminal_reward_function_for_dynamics_env(terminal_func=RandomTerminalFunc(name='random_p'),
                                                           reward_func=RandomRewardFunc('re_fun'))
        return algo, locals()

    def create_eps(self, env_spec):
        return EpsilonGreedy(action_space=env_spec.action_space,
                             init_random_prob=0.5), locals()

    def create_agent(self, algo, env, env_spec, eps=None, name='agent'):
        agent = Agent(env=env,
                      env_spec=env_spec,
                      algo=algo,
                      noise_adder=AgentActionNoiseWrapper(noise=OUNoise(),
                                                          action_weight_scheduler=LinearScheduler(
                                                              t_fn=lambda: get_global_status_collect()(
                                                                  'TOTAL_AGENT_TRAIN_SAMPLE_COUNT'),
                                                              schedule_timesteps=100,
                                                              final_p=1.0,
                                                              initial_p=0.0),
                                                          noise_weight_scheduler=LinearScheduler(
                                                              t_fn=lambda: get_global_status_collect()(
                                                                  'TOTAL_AGENT_TRAIN_SAMPLE_COUNT'),
                                                              schedule_timesteps=100,
                                                              final_p=0.0,
                                                              initial_p=1.0)),
                      name=name,
                      algo_saving_scheduler=PeriodicalEventSchedule(
                          t_fn=lambda: get_global_status_collect()('TOTAL_ENV_STEP_TRAIN_SAMPLE_COUNT'),
                          trigger_every_step=20,
                          after_t=10),
                      exploration_strategy=eps)
        return agent, locals()

    def create_train_test_flow(self, agent, traj_flag=False):

        flow = TrainTestFlow(
            train_sample_count_func=lambda: get_global_status_collect()('TOTAL_ENV_STEP_TRAIN_SAMPLE_COUNT'),
            config_or_config_dict={
                "TEST_EVERY_SAMPLE_COUNT": 10,
                "TRAIN_EVERY_SAMPLE_COUNT": 10,
                "START_TRAIN_AFTER_SAMPLE_COUNT": 5,
                "START_TEST_AFTER_SAMPLE_COUNT": 5,
            },
            func_dict={
                'test': {'func': agent.test,
                         'args': list(),
                         'kwargs': dict(sample_count=1),
                         },
                'train': {'func': agent.train,
                          'args': list(),
                          'kwargs': dict(),
                          },
                'sample': {'func': agent.sample,
                           'args': list(),
                           'kwargs': dict(sample_count=100 if not traj_flag else 1,
                                          env=agent.env,
                                          sample_type='trajectory' if traj_flag else 'transition',
                                          in_which_status='TRAIN',
                                          store_flag=True),
                           },
            }
        )
        return flow

    def create_dyna_flow(self, agent, env):
        flow = DynaFlow(
            train_sample_count_func=lambda: get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT'),
            config_or_config_dict={
                "TEST_ALGO_EVERY_REAL_SAMPLE_COUNT": 10,
                "TEST_DYNAMICS_EVERY_REAL_SAMPLE_COUNT": 10,
                "TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_REAL_ENV": 10,
                "TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_DYNAMICS_ENV": 10,
                "TRAIN_DYNAMICS_EVERY_REAL_SAMPLE_COUNT": 10,
                "START_TRAIN_ALGO_AFTER_SAMPLE_COUNT": 1,
                "START_TRAIN_DYNAMICS_AFTER_SAMPLE_COUNT": 1,
                "START_TEST_ALGO_AFTER_SAMPLE_COUNT": 1,
                "START_TEST_DYNAMICS_AFTER_SAMPLE_COUNT": 1,
                "WARM_UP_DYNAMICS_SAMPLES": 1
            },
            func_dict={
                'train_algo': {'func': agent.train,
                               'args': list(),
                               'kwargs': dict(state='state_agent_training')},
                'train_algo_from_synthesized_data': {'func': agent.train,
                                                     'args': list(),
                                                     'kwargs': dict(state='state_agent_training')},
                'train_dynamics': {'func': agent.train,
                                   'args': list(),
                                   'kwargs': dict(state='state_dynamics_training')},
                'test_algo': {'func': agent.test,
                              'args': list(),
                              'kwargs': dict(sample_count=10)},
                'test_dynamics': {'func': agent.algo.test_dynamics,
                                  'args': list(),
                                  'kwargs': dict(sample_count=10, env=env)},
                'sample_from_real_env': {'func': agent.sample,
                                         'args': list(),
                                         'kwargs': dict(sample_count=10,
                                                        env=agent.env,
                                                        in_which_status='TRAIN',
                                                        store_flag=True)},
                'sample_from_dynamics_env': {'func': agent.sample,
                                             'args': list(),
                                             'kwargs': dict(sample_count=10,
                                                            env=agent.algo.dynamics_env,
                                                            in_which_status='TRAIN',
                                                            store_flag=True)}
            }
        )
        return flow, locals()

    def create_exp(self, name, env, agent, flow=None, traj_flag=False):
        experiment = Experiment(
            tuner=None,
            env=env,
            agent=agent,
            flow=self.create_train_test_flow(agent, traj_flag=traj_flag) if not flow else flow,
            name=name + 'experiment_debug'
        )
        return experiment

    def create_dyna(self, env_spec=None, model_free_algo=None, dyanmics_model=None,
                    name='dyna'):
        if not env_spec:
            model_free_algo, local = self.create_ddpg()
            dyanmics_model, _ = self.create_continuous_mlp_global_dynamics_model(env_spec=local['env_spec'])
            env_spec = local['env_spec']
            env = local['env']
        algo = Dyna(env_spec=env_spec,
                    name=name,
                    model_free_algo=model_free_algo,
                    dynamics_model=dyanmics_model,
                    config_or_config_dict=dict(
                        dynamics_model_train_iter=1,
                        model_free_algo_train_iter=1
                    ))
        algo.set_terminal_reward_function_for_dynamics_env(terminal_func=RandomTerminalFunc(),
                                                           reward_func=RandomRewardFunc())
        return algo, locals()

    def create_continuous_mlp_global_dynamics_model(self, env_spec, name='continuous_mlp_global_dynamics_model'):
        mlp_dyna = ContinuousMLPGlobalDynamicsModel(
            env_spec=env_spec,
            name_scope=name,
            name=name,
            state_input_scaler=RunningStandardScaler(dims=env_spec.flat_obs_dim),
            action_input_scaler=RunningStandardScaler(dims=env_spec.flat_action_dim),
            output_delta_state_scaler=RunningStandardScaler(dims=env_spec.flat_obs_dim),
            learning_rate=0.01,
            mlp_config=[
                {
                    "ACT": "RELU",
                    "B_INIT_VALUE": 0.0,
                    "NAME": "1",
                    "N_UNITS": 16,
                    "TYPE": "DENSE",
                    "W_NORMAL_STDDEV": 0.03
                },
                {
                    "ACT": "LINEAR",
                    "B_INIT_VALUE": 0.0,
                    "NAME": "OUPTUT",
                    "N_UNITS": env_spec.flat_obs_dim,
                    "TYPE": "DENSE",
                    "W_NORMAL_STDDEV": 0.03
                }
            ])
        return mlp_dyna, locals()

    def create_mlp_deterministic_policy(self, env_spec, name='mlp_policy'):
        policy = DeterministicMLPPolicy(env_spec=env_spec,
                                        name=name,
                                        name_scope=name,
                                        mlp_config=[
                                            {
                                                "ACT": "RELU",
                                                "B_INIT_VALUE": 0.0,
                                                "NAME": "1",
                                                "N_UNITS": 16,
                                                "TYPE": "DENSE",
                                                "W_NORMAL_STDDEV": 0.03
                                            },
                                            {
                                                "ACT": "LINEAR",
                                                "B_INIT_VALUE": 0.0,
                                                "NAME": "OUPTUT",
                                                "N_UNITS": env_spec.flat_action_dim,
                                                "TYPE": "DENSE",
                                                "W_NORMAL_STDDEV": 0.03
                                            }
                                        ],
                                        output_high=None,
                                        output_low=None,
                                        output_norm=None,
                                        input_norm=None,
                                        reuse=False)
        return policy, locals()

    def create_uniform_policy(self, env_spec, name='uni_policy'):

        return UniformRandomPolicy(env_spec=env_spec, name=name), locals()

    def create_constant_action_policy(self, env_spec, name='constant_policy'):
        return ConstantActionPolicy(env_spec=env_spec,
                                    name=name,
                                    config_or_config_dict=dict(
                                        ACTION_VALUE=np.array(env_spec.action_space.sample()))), locals()

    def create_ilqr_policy(self, env_id='Pendulum-v0'):

        class DebuggingCostFunc(CostFunc):
            def __call__(self, state=None, action=None, new_state=None, **kwargs) -> float:
                return float(np.sum(action * action))

        env = make(env_id)
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)
        dyna = UniformRandomDynamicsModel(env_spec=env_spec)
        dyna.init()
        policy = iLQRPolicy(env_spec=env_spec,
                            T=50,
                            delta=0.0005,
                            iteration=5,
                            dynamics=dyna,
                            dynamics_model_train_iter=1,
                            cost_fn=DebuggingCostFunc())
        return policy, locals()

    def sample_transition(self, env, count=100):
        data = TransitionData(env.env_spec)
        st = env.get_state()
        for i in range(count):
            ac = env.env_spec.action_space.sample()
            new_st, re, done, info = env.step(action=ac)
            data.append(state=st,
                        action=ac,
                        new_state=new_st,
                        done=done,
                        reward=re)
        return data

    def register_global_status_when_test(self, agent, env):
        """
        this func should be called only for unit test
        :param agent:
        :param env:
        :return:
        """
        get_global_status_collect().register_info_key_status(obj=agent,
                                                             info_key='predict_counter',
                                                             under_status='TRAIN',
                                                             return_name='TOTAL_AGENT_TRAIN_SAMPLE_COUNT')
        get_global_status_collect().register_info_key_status(obj=agent,
                                                             info_key='predict_counter',
                                                             under_status='TEST',
                                                             return_name='TOTAL_AGENT_TEST_SAMPLE_COUNT')
        get_global_status_collect().register_info_key_status(obj=agent,
                                                             info_key='update_counter',
                                                             under_status='TRAIN',
                                                             return_name='TOTAL_AGENT_UPDATE_COUNT')
        get_global_status_collect().register_info_key_status(obj=env,
                                                             info_key='step',
                                                             under_status='TEST',
                                                             return_name='TOTAL_ENV_STEP_TEST_SAMPLE_COUNT')
        get_global_status_collect().register_info_key_status(obj=env,
                                                             info_key='step',
                                                             under_status='TRAIN',
                                                             return_name='TOTAL_ENV_STEP_TRAIN_SAMPLE_COUNT')



"""This file is used for specifying various schedules that evolve over
time throughout the execution of the algorithm, such as:
 - learning rate for the optimizer
 - exploration epsilon for the epsilon greedy exploration strategy
 - beta parameter for beta parameter in prioritized replay

Each schedule has a function `value(t)` which returns the current value
of the parameter given the timestep t of the optimization procedure.
"""
from typeguard import typechecked
from baconian.common.error import *
from baconian.common.logging import ConsoleLogger


class Scheduler(object):
    def __init__(self):
        self.final_p = None
        self.initial_p = None

    def value(self):
        """Value of the schedule at time t"""
        raise NotImplementedError()


class ConstantScheduler(Scheduler):
    def __init__(self, value):
        """Value remains constant over time.

        Parameters
        ----------
        value: float
            Constant value of the schedule
        """
        self._v = value
        Scheduler.__init__(self)

    def value(self):
        """See Schedule.value"""
        return self._v


def linear_interpolation(l, r, alpha):
    return l + alpha * (r - l)


class PiecewiseScheduler(Scheduler):
    def __init__(self, endpoints, t_fn, interpolation=linear_interpolation, outside_value=None):
        """Piecewise schedule.

        endpoints: [(int, int)]
            list of pairs `(time, value)` meanining that schedule should output
            `value` when `t==time`. All the values for time must be sorted in
            an increasing order. When t is between two times, e.g. `(time_a, value_a)`
            and `(time_b, value_b)`, such that `time_a <= t < time_b` then value outputs
            `interpolation(value_a, value_b, alpha)` where alpha is a fraction of
            time passed between `time_a` and `time_b` for time `t`.
        interpolation: lambda float, float, float: float
            a function that takes value to the left and to the right of t according
            to the `endpoints`. Alpha is the fraction of distance from left endpoint to
            right endpoint that t has covered. See linear_interpolation for example.
        outside_value: float
            if the value is requested outside of all the intervals sepecified in
            `endpoints` this value is returned. If None then AssertionError is
            raised when outside value is requested.
        """
        idxes = [e[0] for e in endpoints]
        assert idxes == sorted(idxes)
        Scheduler.__init__(self)

        self._interpolation = interpolation
        self._outside_value = outside_value
        self._endpoints = endpoints
        self.t_fn = t_fn
        assert callable(self.t_fn)

    def value(self):
        """See Schedule.value"""
        t = wrap_t_fn(self.t_fn)
        for (l_t, l), (r_t, r) in zip(self._endpoints[:-1], self._endpoints[1:]):
            if l_t <= t and t < r_t:
                alpha = float(t - l_t) / (r_t - l_t)
                return self._interpolation(l, r, alpha)

        # t does not belong to any of the pieces, so doom.
        assert self._outside_value is not None
        return self._outside_value


class LinearScheduler(Scheduler):
    @typechecked
    def __init__(self, t_fn, schedule_timesteps: int, final_p: float, initial_p=1.0):
        """Linear interpolation between initial_p and final_p over
        schedule_timesteps. After this many timesteps pass final_p is
        returned.

        Parameters
        ----------
        schedule_timesteps: int
            Number of timesteps for which to linearly anneal initial_p
            to final_p
        initial_p: float
            initial output value
        final_p: float
            final output value
        """
        Scheduler.__init__(self)
        self.schedule_timesteps = schedule_timesteps
        self.final_p = final_p
        self.initial_p = initial_p
        self.t_fn = t_fn
        if not callable(self.t_fn):
            raise TypeError("t_fn {} is not callable".format(self.t_fn))

    def value(self):
        t = wrap_t_fn(self.t_fn)
        """See Schedule.value"""
        fraction = min(float(t) / self.schedule_timesteps, 1.0)
        return self.initial_p + fraction * (self.final_p - self.initial_p)


class EventScheduler(Scheduler):
    def value(self) -> bool:
        return False


class PeriodicalEventSchedule(EventScheduler):
    """
    Trigger an event with certain scheduled period
    """

    def __init__(self, t_fn, trigger_every_step, after_t=0):
        super().__init__()
        self.t_fn = t_fn
        self.trigger_every_step = trigger_every_step
        self.after_t = after_t
        self.last_saved_t_value = -1

    def value(self) -> bool:
        """
        return a boolean, true for trigger this event, false for not.
        :return:
        """
        t = wrap_t_fn(self.t_fn)
        if t < self.after_t:
            return False
        else:
            if t - self.last_saved_t_value >= self.trigger_every_step:
                self.last_saved_t_value = t
                return True
            else:
                return False


def wrap_t_fn(t_fn):
    try:
        return t_fn()
    except StatusInfoNotRegisteredError:
        ConsoleLogger().print('error', 'StatusInfoNotRegisteredError occurred, return with 0')
        return 0



from baconian.config.global_config import GlobalConfig
from baconian.test.tests.set_up.setup import BaseTestCase
from baconian.common.error import *
from baconian.algo.dynamics.reward_func.reward_func import RandomRewardFunc
from baconian.core.core import Basic

class TestCore(BaseTestCase):
    def test_global_config(self):
        GlobalConfig().set_new_config(config_dict=dict(DEFAULT_BASIC_INIT_STATUS='test'))
        assert GlobalConfig().DEFAULT_BASIC_INIT_STATUS == 'test'

        GlobalConfig().freeze_flag = True
        try:
            GlobalConfig().set_new_config(config_dict=dict(DEFAULT_BASIC_INIT_STATUS='test'))
        except AttemptToChangeFreezeGlobalConfigError:
            pass
        else:
            raise TypeError

        try:
            GlobalConfig().set('DEFAULT_LOG_PATH', 'tmp')
        except AttemptToChangeFreezeGlobalConfigError:
            pass
        else:
            raise TypeError

        try:
            GlobalConfig().DEFAULT_LOG_PATH = 'tmp'
        except AttemptToChangeFreezeGlobalConfigError:
            pass
        else:
            raise TypeError
        GlobalConfig().unfreeze()

    def test_config(self):
        config, _ = self.create_dict_config()

    def test_name_register(self):
        a = RandomRewardFunc()
        self.assertTrue(a.allow_duplicate_name)
        b = RandomRewardFunc()
        a = Basic(name='s')
        try:
            b = Basic(name='s')
        except GlobalNameExistedError as e:
            pass
        else:
            raise NotCatchCorrectExceptionError()



import abc
from baconian.common.logging import ConsoleLogger
from baconian.common.error import *
from functools import wraps
import numpy as np

__all__ = ['Status', 'StatusWithSubInfo', 'StatusWithSingleInfo', 'StatusWithInfo', 'StatusCollector',
           'reset_global_status_collect', 'register_counter_info_to_status_decorator', 'get_global_status_collect']


class Status(object):
    """
    One of the core module, a class to indicate the current status of an object in Baconian.
    """

    def __init__(self, obj):
        """
        The object that status will composed into.
        :param obj: A object of Baconian, e.g., Environment, Agent, Algo.
        """
        self.obj = obj
        self._status_val = None
        if hasattr(obj, 'STATUS_LIST'):
            self._status_list = obj.STATUS_LIST
        else:
            self._status_list = None
        if hasattr(obj, 'INIT_STATUS') and obj.INIT_STATUS is not None:
            self.set_status(new_status=obj.INIT_STATUS)
        else:
            self._status_val = None

    def __call__(self, *args, **kwargs):
        return dict(status=self._status_val)

    def set_status(self, new_status: str):
        if not isinstance(new_status, str):
            raise TypeError("{} is not string".format(new_status))
        if self._status_list:
            try:
                assert new_status in self._status_list
            except AssertionError as e:
                ConsoleLogger().print('info', "{} New status :{} not in the status list: {} ".format(e, new_status,
                                                                                                     self._status_list))
            self._status_val = new_status
        else:
            self._status_val = new_status

    def get_status(self) -> dict:
        return self()


class StatusWithInfo(Status):
    @abc.abstractmethod
    def append_new_info(self, *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def has_info(self, *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def update_info(self, *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def reset(self):
        raise NotImplementedError

    @abc.abstractmethod
    def get_specific_info_key_status(self, info_key, *args, **kwargs):
        raise NotImplementedError


class StatusWithSingleInfo(StatusWithInfo):
    def __init__(self, obj):
        super().__init__(obj)

        self._info_dict = {}

    def __call__(self, *args, **kwargs):
        res = super().__call__(*args, **kwargs)
        return {**res, **self._info_dict}

    def set_status(self, new_status: str):
        return super().set_status(new_status)

    def get_status(self) -> dict:
        return self()

    def append_new_info(self, info_key: str, init_value, under_status=None):
        if info_key == 'status':
            raise ValueError("can use key: status which is a system default key")
        if info_key in self._info_dict:
            return
        else:
            self._info_dict[info_key] = init_value

    def get_specific_info_key_status(self, info_key, *args, **kwargs):
        try:
            return self._info_dict[info_key]
        except KeyError:
            ConsoleLogger().print('ERROR',
                                  'try to access info key status: {} of obj: {}'.format(info_key, self.obj.name))
            return None

    def has_info(self, info_key, **kwargs):
        return info_key in self._info_dict

    def update_info(self, info_key, increment, under_status=None):
        if not self.has_info(info_key=info_key):
            self.append_new_info(info_key=info_key, init_value=0)
        self._info_dict[info_key] += increment

    def reset(self):
        self._info_dict = {}


class StatusWithSubInfo(StatusWithInfo):

    def __init__(self, obj):
        super().__init__(obj)
        if not hasattr(obj, 'STATUS_LIST') or not hasattr(obj, 'INIT_STATUS'):
            raise ValueError(
                "StatusWithSubInfo require the source object to have class attr: STATUS_LIST and INIT_STATUS")

        self._info_dict_with_sub_info = {}
        for key in self._status_list:
            self._info_dict_with_sub_info[key] = {}

    def __call__(self, *args, **kwargs):
        res = super().__call__(*args, **kwargs)
        return {**res, **self._info_dict_with_sub_info[self._status_val]}

    def set_status(self, new_status: str):
        return super().set_status(new_status)

    def get_status(self) -> dict:
        return self()

    def get_specific_info_key_status(self, info_key, under_status, *args, **kwargs):
        res = self._get_specific_info_key_status(info_key=info_key, under_status=under_status, *args, **kwargs)
        if res is None:
            ConsoleLogger().print('error', 'try to access info key status: {} under status {} of obj: {}'.
                                  format(info_key, under_status, self.obj.name))
        else:
            return res

    def _get_specific_info_key_status(self, info_key, under_status, *args, **kwargs):
        try:
            return self._info_dict_with_sub_info[under_status][info_key]
        except KeyError:
            return None

    def group_specific_info_key(self, info_key, group_way):
        assert group_way in ('sum', 'max', 'min', 'mean')
        res = []
        for st in self._status_list:
            re = self._get_specific_info_key_status(info_key=info_key, under_status=st)
            if re:
                res.append(re)
        if group_way == 'sum':
            return sum(res)
        if group_way == 'max':
            return max(res)
        if group_way == 'min':
            return min(res)
        if group_way == 'mean':
            return np.mean(np.array(res)).item()

    def append_new_info(self, info_key: str, init_value, under_status=None):
        if not under_status:
            under_status = self._status_val
        if info_key == 'status':
            raise ValueError("can use key: status which is a system default key")
        if info_key in self._info_dict_with_sub_info[under_status]:
            return
        else:
            self._info_dict_with_sub_info[under_status][info_key] = init_value

    def has_info(self, info_key, under_status=None):
        if not under_status:
            under_status = self._status_val
        return info_key in self._info_dict_with_sub_info[under_status]

    def update_info(self, info_key, increment, under_status=None):
        if not under_status:
            under_status = self._status_val
        if not self.has_info(info_key=info_key, under_status=under_status):
            self.append_new_info(info_key=info_key, init_value=0, under_status=under_status)
        self._info_dict_with_sub_info[under_status][info_key] += increment

    def reset(self):
        for key in self._status_list:
            self._info_dict_with_sub_info[key] = {}


class StatusCollector(object):

    def __init__(self):
        self._register_status_dict = []

    def __call__(self, key: str = None, *args, **kwargs):
        if key:
            for val in self._register_status_dict:
                if val['return_name'] == key:
                    obj = val['obj']
                    assert hasattr(obj, '_status')
                    assert isinstance(getattr(obj, '_status'), StatusWithInfo)
                    if getattr(obj, '_status').has_info(info_key=val['info_key'],
                                                        under_status=val['under_status']) is False:
                        raise StatusInfoNotRegisteredError(
                            '{} do not have {} under {}'.format(obj, val['info_key'], val['under_status']))

                    res = obj._status.get_specific_info_key_status(under_status=val['under_status'],
                                                                   info_key=val['info_key'])
                    return res
        else:
            stat_dict = dict()
            for val in self._register_status_dict:
                obj = val['obj']
                assert hasattr(obj, '_status')
                assert isinstance(getattr(obj, '_status'), StatusWithInfo)
                if getattr(obj, '_status').has_info(info_key=val['info_key'],
                                                    under_status=val['under_status']) is False:
                    raise StatusInfoNotRegisteredError(
                        '{} do not have {} under {}'.format(obj, val['info_key'], val['under_status']))

                res = obj._status.get_specific_info_key_status(under_status=val['under_status'],
                                                               info_key=val['info_key'])
                stat_dict[val['return_name']] = res
            return stat_dict

    def get_status(self) -> dict:
        return self()

    def register_info_key_status(self, obj, info_key: str, return_name: str, under_status=None):
        ConsoleLogger().print('info',
                              'registered obj: {}, key: {}, return name: {}, under status: {}'.format(obj, info_key,
                                                                                                      return_name,
                                                                                                      under_status))
        for val in self._register_status_dict:
            assert return_name != val['return_name']
        self._register_status_dict.append(
            dict(obj=obj, info_key=info_key, under_status=under_status, return_name=return_name))
        try:
            self(info_key)
        except StatusInfoNotRegisteredError as e:
            ConsoleLogger().print('warning',
                                  'new registred info: obj: {}, key: {}, return name: {}, under status: {} can not be detected now'.format(
                                      obj, info_key,
                                      return_name,
                                      under_status))

    def reset(self):
        self._register_status_dict = []


def register_counter_info_to_status_decorator(increment, info_key, under_status: (str, tuple) = None,
                                              ignore_wrong_status=False):
    def wrap(fn):
        if under_status:
            assert isinstance(under_status, (str, tuple))
            if isinstance(under_status, str):
                final_st = tuple([under_status])
            else:
                final_st = under_status

        else:
            final_st = (None,)

        @wraps(fn)
        def wrap_with_self(self, *args, **kwargs):
            # todo record() called in fn will lost the just appended info_key at the very first
            obj = self
            if not hasattr(obj, '_status') or not isinstance(getattr(obj, '_status'), StatusWithInfo):
                raise ValueError(
                    ' the object {} does not not have attribute StatusWithInfo instance or hold wrong type of Status'.format(
                        obj))

            assert isinstance(getattr(obj, '_status'), StatusWithInfo)
            obj_status = getattr(obj, '_status')
            for st in final_st:
                obj_status.append_new_info(info_key=info_key, init_value=0, under_status=st)
            res = fn(self, *args, **kwargs)
            for st in final_st:
                if st and st != obj.get_status()['status'] and not ignore_wrong_status:
                    raise ValueError('register counter info under status: {} but got status {}'.format(st,
                                                                                                       obj.get_status()[
                                                                                                           'status']))
            obj_status.update_info(info_key=info_key, increment=increment,
                                   under_status=obj.get_status()['status'])
            return res

        return wrap_with_self

    return wrap


# _global_experiment_status = StatusWithSingleInfo(obj=None)
#
# from baconian.config.global_config import GlobalConfig
#
# for key in GlobalConfig().DEFAULT_EXPERIMENT_END_POINT.keys():
#     _global_experiment_status.append_new_info(info_key=key, init_value=0)

# def get_global_experiment_status() -> StatusWithSingleInfo:
#     return globals()['_global_experiment_status']
#

# def reset_global_experiment_status():
#     globals()['_global_experiment_status'].reset()
#

_global_status_collector = StatusCollector()


def get_global_status_collect() -> StatusCollector:
    return globals()['_global_status_collector']


def reset_global_status_collect():
    globals()['_global_status_collector'].reset()



from baconian.common.special import flatten_n
from baconian.core.core import EnvSpec
from baconian.algo.dynamics.dynamics_model import GlobalDynamicsModel, DifferentiableDynamics, TrainableDyanmicsModel
import tensorflow as tf
from baconian.tf.tf_parameters import ParametersWithTensorflowVariable
from baconian.tf.mlp import MLP
from baconian.common.sampler.sample_data import TransitionData
from typeguard import typechecked
import numpy as np
from baconian.tf.util import *
from baconian.algo.misc.placeholder_input import PlaceholderInput
import overrides
from baconian.common.error import *
from baconian.common.logging import record_return_decorator
from baconian.core.status import register_counter_info_to_status_decorator, StatusWithSubInfo
from baconian.common.spaces.box import Box
from baconian.common.data_pre_processing import DataScaler, IdenticalDataScaler


class ContinuousMLPGlobalDynamicsModel(GlobalDynamicsModel, DifferentiableDynamics, PlaceholderInput,
                                       TrainableDyanmicsModel):
    STATUS_LIST = GlobalDynamicsModel.STATUS_LIST + ('TRAIN',)
    INIT_STATUS = 'CREATED'

    def __init__(self, env_spec: EnvSpec,
                 name_scope: str,
                 name: str,
                 mlp_config: list,
                 learning_rate: float,
                 state_input_scaler: DataScaler = None,
                 action_input_scaler: DataScaler = None,
                 output_delta_state_scaler: DataScaler = None,
                 init_state=None):
        if not isinstance(env_spec.obs_space, Box):
            raise TypeError('ContinuousMLPGlobalDynamicsModel only support to predict state that hold space Box type')
        GlobalDynamicsModel.__init__(self,
                                     env_spec=env_spec,
                                     parameters=None,
                                     name=name,
                                     state_input_scaler=state_input_scaler,
                                     action_input_scaler=action_input_scaler,
                                     init_state=init_state)
        with tf.variable_scope(name_scope):
            state_input = tf.placeholder(shape=[None, env_spec.flat_obs_dim], dtype=tf.float32, name='state_ph')
            action_input = tf.placeholder(shape=[None, env_spec.flat_action_dim], dtype=tf.float32,
                                          name='action_ph')
            mlp_input_ph = tf.concat([state_input, action_input], axis=1, name='state_action_input')
            delta_state_label_ph = tf.placeholder(shape=[None, env_spec.flat_obs_dim], dtype=tf.float32,
                                                  name='delta_state_label_ph')
        mlp_net = MLP(input_ph=mlp_input_ph,
                      reuse=False,
                      mlp_config=mlp_config,
                      name_scope=name_scope,
                      net_name='mlp')
        if mlp_net.output.shape[1] != env_spec.flat_obs_dim:
            raise InappropriateParameterSetting(
                "mlp output dims {} != env spec obs dim {}".format(mlp_net.output.shape[1], env_spec.flat_obs_dim))

        parameters = ParametersWithTensorflowVariable(tf_var_list=mlp_net.var_list,
                                                      name=name + '_''mlp_continuous_dynamics_model',
                                                      rest_parameters=dict(learning_rate=learning_rate))

        DifferentiableDynamics.__init__(self,
                                        input_node_dict=dict(state_input=state_input,
                                                             action_action_input=action_input),
                                        output_node_dict=dict(delta_state_output=mlp_net.output))
        PlaceholderInput.__init__(self, parameters=parameters)

        self.mlp_config = mlp_config
        self.name_scope = name_scope
        self.action_input = action_input
        self.state_input = state_input
        self.mlp_input_ph = mlp_input_ph
        self.delta_state_label_ph = delta_state_label_ph
        self.delta_state_output = mlp_net.output
        self.mlp_net = mlp_net
        self.output_delta_state_scaler = output_delta_state_scaler if output_delta_state_scaler else IdenticalDataScaler(
            dims=self.env_spec.flat_obs_dim)
        self._status = StatusWithSubInfo(obj=self)

        with tf.variable_scope(name_scope):
            with tf.variable_scope('train'):
                self.loss, self.optimizer, self.optimize_op = self._setup_loss()
        train_var_list = get_tf_collection_var_list(key=tf.GraphKeys.GLOBAL_VARIABLES,
                                                    scope='{}/train'.format(
                                                        name_scope)) + self.optimizer.variables()

        self.parameters.set_tf_var_list(sorted(list(set(train_var_list)), key=lambda x: x.name))

    def init(self, source_obj=None):
        self.parameters.init()
        if source_obj:
            self.copy_from(obj=source_obj)
        GlobalDynamicsModel.init(self)

    @register_counter_info_to_status_decorator(increment=1, info_key='step')
    def step(self, action: np.ndarray, state=None, **kwargs_for_transit):
        return super().step(action, state, **kwargs_for_transit)

    @record_return_decorator(which_recorder='self')
    @register_counter_info_to_status_decorator(increment=1, info_key='train_counter', under_status='TRAIN')
    def train(self, batch_data: TransitionData, **kwargs) -> dict:
        self.set_status('TRAIN')

        self.state_input_scaler.update_scaler(batch_data.state_set)
        self.action_input_scaler.update_scaler(batch_data.action_set)
        self.output_delta_state_scaler.update_scaler(batch_data.new_state_set - batch_data.state_set)

        tf_sess = kwargs['sess'] if ('sess' in kwargs and kwargs['sess']) else tf.get_default_session()
        train_iter = self.parameters('train_iter') if 'train_iter' not in kwargs else kwargs['train_iter']
        feed_dict = {
            self.state_input: self.state_input_scaler.process(batch_data.state_set),
            self.action_input: self.action_input_scaler.process(
                flatten_n(self.env_spec.action_space, batch_data.action_set)),
            self.delta_state_label_ph: self.output_delta_state_scaler.process(
                batch_data.new_state_set - batch_data.state_set),
            **self.parameters.return_tf_parameter_feed_dict()
        }
        average_loss = 0.0

        for i in range(train_iter):
            loss, _ = tf_sess.run([self.loss, self.optimize_op],
                                  feed_dict=feed_dict)
            average_loss += loss
        return dict(average_loss=average_loss / train_iter)

    def save(self, *args, **kwargs):
        return PlaceholderInput.save(self, *args, **kwargs)

    def load(self, *args, **kwargs):
        return PlaceholderInput.load(self, *args, **kwargs)

    @overrides.overrides
    def copy_from(self, obj: PlaceholderInput) -> bool:
        return PlaceholderInput.copy_from(self, obj)

    def _state_transit(self, state, action, **kwargs) -> np.ndarray:
        state = self.state_input_scaler.process(
            np.array(state).reshape(self.env_spec.obs_shape)) if state is not None else self.state
        action = self.action_input_scaler.process(self.env_spec.flat_action(action))

        if 'sess' in kwargs:
            tf_sess = kwargs['sess']
        else:
            tf_sess = tf.get_default_session()

        if len(state.shape) < 2:
            state = np.expand_dims(state, 0)
        if len(action.shape) < 2:
            action = np.expand_dims(action, 0)
        delta_state = tf_sess.run(self.delta_state_output,
                                  feed_dict={
                                      self.action_input: action,
                                      self.state_input: state
                                  })
        new_state = np.clip(
            np.squeeze(self.output_delta_state_scaler.inverse_process(data=np.squeeze(delta_state)) + state),
            self.env_spec.obs_space.low, self.env_spec.obs_space.high)
        return new_state

    def _setup_loss(self):
        reg_loss = tf.get_collection(tf.GraphKeys.REGULARIZATION_LOSSES, scope=self.name_scope)
        loss = tf.reduce_sum((self.mlp_net.output - self.delta_state_label_ph) ** 2)
        if len(reg_loss) > 0:
            loss += tf.reduce_sum(reg_loss)
        optimizer = tf.train.AdamOptimizer(learning_rate=self.parameters('learning_rate'))
        optimize_op = optimizer.minimize(loss=loss, var_list=self.parameters('tf_var_list'))
        return loss, optimizer, optimize_op
